/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Bézier curve interpolation for spike detection
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "BezierInterpolation.h"

/*!*****************************************************************************
 * @brief Solution of the Bézier curve derivative
 *
 * Compute value t such that B'(t) = a.t + b = 0
 *
 * @param a Factor of the Bezier curve derivative
 * @param b Constant term in the Bezier curve derivative
 * @param t Solution of the Bezier curve derivative
 *
 * @result Number of solutions (-1 for infinity)
 ******************************************************************************/
int BezierDerivative(double a, double b, double *t)
{        
  if(a != 0){     // One solution 
    *t = - b / a;
    return 1;
  }else{
    if(b == 0){   // Infinite number of solutions
      return -1; 
    }else{        // No solution
      return 0;
    }
  }
}

/*!*****************************************************************************
 * @brief Value of quadratic Bézier curve defined by (P1, P2, P3), for parameter value t
 *
 * Compute the value of B(t) for the given t.
 *
 * @param t        Parameter value of the Bézier curve
 * @param P1,P2,P3 The 3 control points that define the Bézier curve (along one axis: x or y)
 *
 * @result Value of the curve for parameter t
 ******************************************************************************/
double BezierValue(double t, double P1, double P2, double P3)
{
  double omt = 1 - t;

  return omt * omt * P1 + 2 * omt * t * P2 + t * t * P3;
}

/*!*****************************************************************************
 * @brief Computes the extremum of the Bézier curve defined by (P1, P2, P3)
 * along the specified dimension
 *
 * Quadratic Bézier curve B(t) = (1-t)².P1 + 2.(1-t).P2 + t².P3,  t in [0,1]
 * where points Pi(xi, yi) are defined by 2D coordinates x and y.
 *
 * @param Axis     The coordinate to be maximized ('x' or 'y')
 * @param P1,P2,P3 The three control points of the Bézier curve 
 *
 * @result 2D point corresponding to the extremum along the specified Axis 
 ******************************************************************************/
Point2D BezierMax(char Axis, Point2D *P1, Point2D *P2, Point2D *P3)
{
  Point2D Sol;            // Max point of the Bézier curve along axis Axis
  int nbSols;             // Number of solutions
  double a, b;            // Parameters of the Bezier curve derivative
  double Tmpt;            // Temporary solution of Bézier parameter t in [0,1]
  double PS = 0, tS = -1; // Solution of the Bézier curve: parameter tS and point PS
  double cP1, cP2, cP3;   // Coordinates of the Bezier control points P1, P2, P3 along the Axis axis

  // Retrieving of control points coordinates along Axis ('x' or 'y')
  if(Axis == 'x'){
    cP1 = P1->x;
    cP2 = P2->x;
    cP3 = P3->x;
  }else if(Axis == 'y'){
    cP1 = P1->y;
    cP2 = P2->y;
    cP3 = P3->y;
  }else{
    fprintf(stderr, "!!! ERROR: undefined axis specification in Bézier max determination !!!\n");
    exit(EXIT_FAILURE);
  }

  // Computation of the Bezier derivative parameters: B'(t) = a t + b
  a = 2 * (cP1 - 2 * cP2 + cP3); 
  b = 2 * (cP2 - cP1); 

  // Computation of parameter t to obtain a null derivative: B'(Tmpt) = 0, <=> maximum of the Bézier curve
  nbSols = BezierDerivative(a, b, &Tmpt);

  // Computation of the solutions inside range [0,1]
  if(nbSols > 0){
    if(0 <= Tmpt && Tmpt <= 1){
      PS = BezierValue(Tmpt, cP1, cP2, cP3);
      tS = Tmpt;
    }else{
      nbSols--;
    }
  }

  // If there is no solution or the point PS is below cP1, the solution is set to B(0)=P1
  if(nbSols == 0 || cP1 > PS){
    PS = cP1;
    tS = 0;
  }
  // If the point PS is below cP3, the solution is set to B(1)=P3
  if(cP3 > PS){
    PS = cP3;
    tS  = 1;
  }

  // Construction of the final 2D point B(tS)
  if(Axis == 'x'){
    Sol.x = PS;
    Sol.y = BezierValue(tS, P1->y, P2->y, P3->y);
  }else{
    Sol.x = BezierValue(tS, P1->x, P2->x, P3->x);
    Sol.y = PS;
  }

  // Result
  return Sol;
}
