################################################################################
# SiReNe software version 1.0
# Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
#
# Authors : N. AZEVEDO CARVALHO and L. BUHRY
# Years   : 2018-2020
#
# License : GNU General Public License, version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free  Software Foundation; either  version 2 of the  License, or
# (at your option) any later version.
#
# This program is distributed in the  hope that it will be useful, but
# WITHOUT  ANY   WARRANTY;  without  even  the   implied  warranty  of
# MERCHANTABILITY or  FITNESS FOR A  PARTICULAR PURPOSE.  See  the GNU
# General Public License for more details.
#
################################################################################
# File description:
#
# Python file corresponding to the BRIAN simulator version
# of the MSN network example given in TestMSN.c
#
# McCarthy et al 2011
# https://www.pnas.org/content/pnas/108/28/11620.full.pdf
################################################################################
 
from __future__ import division
from brian2 import *
import os
import time
import random
import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
from matplotlib import cm as matpl
import array as arr
from guppy import hpy

# Start time of the initialization
beginInit = time.time()

################################################################################
# PRNG seed initialization
################################################################################
random.seed(a=0)

################################################################################
# Uniform Random Number with given amplitude
#
# The generated number will be between -ampli and +ampli
#
# @param ampli Amplitude of the noise
################################################################################
def unif1(ampli):
   return 1.0 + ampli * (2 * np.random.rand(nbNeurons)- 1)

################################################################################
# Specific noise with given amplitude
#
# The generated number will be between 1-ampli and 1+ampli
#
# @param ampli Amplitude of the noise
################################################################################
def unif2(ampli):
   return 1.0 + ampli * (2 * np.random.rand(nbNeurons, nbNeurons)- 1)

################################################################################
# Simulator Initialization
################################################################################
Time            = 4000    # Simulation run time in ms
defaultclock.dt = 0.01*ms # Time step

################################################################################
# MSN neurons Model
################################################################################
nbNeurons = 100        # network size

Cm  = 1*uF             # Membrane capacitance (in µF)
             
# Conductances of the different currents
gL  = 0.1*msiemens     # Leak conductance (in mS)
gNa = 100.0*msiemens   # Na conductance (in mS)
gK  = 80.0*msiemens    # K conductance (in mS)
gM  = 1.34*msiemens    # M conductance (Normal Condition = 1.34, Parkinson Condition = 1.1) (in mS)

# Ionic reversal Potentials
El  = -67.0*mV         # Ionic reversal Potential of leak current (in mV)
ENa = 50.0*mV          # Ionic reversal Potential of Na current (in mV)
EK  = -100.0*mV        # Ionic reversal Potential of K current (in mV)

### Input Current ###
# Noisy Step 
ampli   = 3                                       # Noise amplitude
tt      = int((Time*ms)/defaultclock.dt+1)        # Number of time steps during the simulation
IappTMP = np.zeros((tt,nbNeurons))                # Input current initialization 
tstep   = 500                                     # Time of the step
for ii in arange(0,tt):
   if ii < (tstep*ms)/defaultclock.dt:
      IappTMP[ii][:] = -10.0 * unif1(ampli)       # Input current before the step
   else:
      IappTMP[ii][:] = 1.19 * unif1(ampli)        # Tnput current after the step
   
Iapp = TimedArray(IappTMP*uA, dt=defaultclock.dt) # Input current

### Equations ###
eqs = Equations("""
# Voltage derivative
dV/dt = (- IL - INa - IK - IM - Isyni + Iapp(t,i))/Cm: volt

# Currents
IL  = gL * (V - El) : amp              # Leak current
INa = gNa * m**3 * h * (V - ENa) : amp # Na current
IK  = gK * n**4 * (V - EK) :amp        # K current
IM  = gM * p * (V - EK) :amp           # M current

# Gating Variables Derivatives
dm/dt = ( (mInf - m) / mTau) : 1
dh/dt = ( (hInf - h) / hTau) : 1
dn/dt = ( (nInf - n) / nTau) : 1
dp/dt = ( (pInf - p) / pTau) : 1

### Na Current ###
alpham =  ( (0.32/mV * (V + (54.0*mV))) / (1.0 - exp((- (V + (54.0*mV))) / (4.0*mV)))) / ms : Hz
betam  = ( (0.28/mV * (V + (27.0*mV))) / (exp((V + (27.0*mV)) / (5.0*mV)) - 1.0)) / ms : Hz
alphah = (0.128 * exp(-(V + (50.0*mV)) / (18.0*mV))) / ms : Hz
betah  = (4.0 / (1.0 + exp(-(V + (27.0*mV)) / (5.0*mV)))) / ms : Hz
mInf   = (alpham / (alpham + betam)) : 1
hInf   = (alphah / (alphah + betah)) : 1
mTau   = (1 / (alpham + betam)) : second
hTau   = (1 / (alphah + betah)) : second

### K Current ###
alphan = ( (0.032/mV * (V + (52.0*mV))) / (1.0 - exp( -(V + (52.0*mV)) / (5.0*mV)))) / ms : Hz
betan  = (0.5 * exp( -(V + (57.0* mV)) / (40.0*mV))) /ms : Hz
nInf   = alphan / (alphan + betan) : 1
nTau   = (1 / (alphan + betan)) : second

### M Current ###
alphap = ( (0.0003209/mV *  (V + (30.0*mV))) / (1 - exp( -(V + (30.0*mV)) / (9.0*mV)))) / ms : Hz
betap  = ( (-0.0003209/mV * (V + (30.0*mV))) / (1 - exp( (V + (30.0*mV)) / (9.0*mV)))) / ms : Hz
pInf   = (alphap / (alphap + betap)) : 1
pTau   = (1.0 / (alphap + betap)): second

### Input Current ###
Iappm = Iapp(t,i) : amp 

### Synapses ###
Isyni  = gi * (V - Esyn) : amp       # Synaptic current
dgi/dt = -gi * (1./taui) : siemens   # Synaptic conductance
""")

### MSN neurons group ###
P = NeuronGroup(nbNeurons, model=eqs, threshold='V>0*mV', refractory=2.5*ms, method='rk2')

### Variables Initialization ###
ampli = 0.1                          # Noise amplitude
P.V   = -63.826  * unif1(ampli) * mV # Initial membrane potential of the neuron (in mV)
P.m   = 0.027781 * unif1(ampli)      # Activation gating variable of Na current (in 1)
P.h   = 0.990915 * unif1(ampli)      # Inactivation gating variable of Na current (in 1)
P.n   = 0.062045 * unif1(ampli)      # Activation gating variable of K current (in 1)
P.p   = 0.022788 * unif1(ampli)      # Activation gating variable of M current (in 1)

whichtorecord = [0, 50, 99]          # Record the number of spikes and a few traces
# whichtorecord = True               # Uncomment to record traces for all neurons

Spikemon = SpikeMonitor(P)           # Get the spike times
# traceV   = StateMonitor(P, 'V', record=whichtorecord) # recording of V, membrane voltage

################################################################################
# Synapse Models
################################################################################
Esyn = -80*mV     # Reversal potential of GABA synapses
taui = 15*msecond # Synaptic time constant (in ms)

### Connections ###
# Connection Matrix
WeigthMatrix = zeros((nbNeurons, nbNeurons))     # Creating a weight matrix
WeigthMatrix = 0.25*mS/nbNeurons * unif2(0.71)   # Random conductances between 0.1 and 0.6, different for all neurons
for j in range(nbNeurons):
   WeigthMatrix[j][j] = 0                        # Neuron is not connected to itself

# Connection Matrix With 30% Connectivity
# percentage = 0.7                               # 30% of connection and 70% of no connection
# NoConni = zeros(int(nbNeurons*percentage))     # Matrix of the non-connected neurons (lines)
# NoConnj = zeros(int(nbNeurons*percentage))     # Matrix of the non-connected neurons (columns) 

# for j in range(int(nbNeurons*percentage)):    
#    NoConni[j] = np.random.randint(0,nbNeurons) # Generating the non-connected i-th line
#    NoConnj[j] = np.random.randint(0,nbNeurons) # Generating the non-connected j-th column

# for j1 in NoConni:
#    for j2 in NoConnj:
#       WeigthMatrix[int(j1)][int(j2)] = 0       # Attributing 0 to the non-connected synapses 
   
Pi = P                                                # Group of inhibitory neurons 
Ci = Synapses(Pi, P,'wi: siemens', on_pre='gi += wi') # In our case, we have only inhibitory neurons
Ci.connect()                                          # Connection probability between Pi and P
Ci.wi = WeigthMatrix.flatten()

################################################################################
# Simulation Run
################################################################################

### Simulation Start ###
begin = time.time() # Simulation start time
run(Time * ms)

#### Memory Profiling ####
h = hpy()

print(h.heap())

### Simulation End ###
end = time.time()   # Simulation end time

print('Simulation time with initialization:', (end-beginInit), 's')
print('Simulation time:', (end-begin), 's')

### Spiking Rate ###
mfreqnbN = len(Spikemon.i)/(4*nbNeurons)
print("The average MSN spiking rate is ", mfreqnbN, " Hz.")

################################################################################
# Different Plots
################################################################################

### Raster Plot ###
figure()
fig2 = matplotlib.pyplot.gcf()
plot(Spikemon.t/ms, Spikemon.i, '.')
xlabel('Time (ms)')
ylabel('Neuron index');

### Potential Plot ###
# figure()
# fig = matplotlib.pyplot.gcf()
# plot(traceV.t/ms, traceV[0].V/mV)
# plot(traceV.t/ms, traceV[50].V/mV)
# plot(traceV.t/ms, traceV[99].V/mV)
# xlabel('t (ms)')
# ylabel('V (mV)')

show()
################################################################################
################################################################################
