/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Display functions for neurons and models
 ******************************************************************************/

#include "Displays.h"

/*!*****************************************************************************
 * @brief Display of a neural model
 *
 * Displayed information:
 *   - Neural model number
 *   - Number of neurons in the model 
 *   - Local neuron numbering
 *   - Number of variables
 *   - Number of parameters
 *   - Parameters of the model
 *   - Spike threshold
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param Output     Output stream of the display
 ******************************************************************************/
void NeuralModelPrint(Simul *Simulation, int ModelNum, FILE *Output)
{
  fprintf(Output, "Neural model: %d\n", ModelNum);
  fprintf(Output, "\tNumber of neurons      = %d\n", Simulation->NeuralModels[ModelNum].NumberOfNeurons);
  fprintf(Output, "\tGlobal neurons numbers = ");

  // Display of neurons numbers
  for(int NeuronNum = 0; NeuronNum < Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
    fprintf(Output, "%d ", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron);
  }
  fprintf(Output, "\n");
  fprintf(Output, "\tNumber of variables    = %d\n", Simulation->NeuralModels[ModelNum].NumberOfVariables);
  fprintf(Output, "\tNumber of parameters   = %d\n", Simulation->NeuralModels[ModelNum].NumberOfParameters);
  fprintf(Output, "\tParameters             = ");

  // Display of model parameters
  for(int p = 0; p < Simulation->NeuralModels[ModelNum].NumberOfParameters; p++){
    fprintf(Output, "%12.6f ", Simulation->NeuralModels[ModelNum].Parameters[p]);
  }
  fprintf(Output, "\n");
  fprintf(Output, "\tSpike treshold         = %.5f\n", Simulation->NeuralModels[ModelNum].SpikeTreshold);
  fprintf(Output, "\n");
}

/*!*****************************************************************************
 * @brief Display of a synaptic model between two neural models
 *
 * Displayed information:
 *   - Synapse model from neuron model SrcModel to neuron model DstModel
 *   - Number of variables 
 *   - Number of parameters
 *   - Parameters
 *
 * @param Simulation Structure where all simulation information is stored
 * @param SrcModel   Source neural model
 * @param DstModel   Destination neural model
 * @param Output     Output stream of the display
 ******************************************************************************/
void SynapseModelPrint(Simul *Simulation, int SrcModel, int DstModel, FILE *Output)
{
  fprintf(Output, "Synaptic model: from %d to %d\n", SrcModel, DstModel);
  if(Simulation->SynapseModels[SrcModel][DstModel].NumberOfParameters >= 0){
    fprintf(Output, "\tNumber of variables  = %d\n", Simulation->SynapseModels[SrcModel][DstModel].NumberOfVariables);
    fprintf(Output, "\tNumber of parameters = %d\n", Simulation->SynapseModels[SrcModel][DstModel].NumberOfParameters);
    fprintf(Output, "\tParameters           = ");

    // Display of model parameters
    for(int p = 0; p < Simulation->SynapseModels[SrcModel][DstModel].NumberOfParameters; p++){
      fprintf(Output, "%12.6f ", Simulation->SynapseModels[SrcModel][DstModel].Parameters[p]);
    }
  }else{
    fprintf(Output, "\tUndefined model\n");
  } 
  fprintf(Output, "\n");
}

/*!*****************************************************************************
 * @brief Display of neuron NeuronNum in model ModelNum
 *
 * Displayed information:
 *   - Number of parameters 
 *   - Parameters 
 *   - Number of variables 
 *   - Initial values 
 *   - Potential index in the Variables array 
 *   - Number of spikes
 *   - Last spike time
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param Output     Output stream of the display
 ******************************************************************************/
void NeuronPrint(Simul *Simulation, int ModelNum, int NeuronNum, FILE *Output)
{
  fprintf(Output, "Neuron: %d\n", NeuronNum);
  fprintf(Output, "\tModel                  = %d\n", ModelNum);
  
  fprintf(Output, "\tNumber of parameters   = %d\n", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfParameters);
  fprintf(Output, "\n");
  fprintf(Output, "\tParameters             = ");

  // Display of neuron parameters
  for(int p = 0; p < Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfParameters; p++){
    fprintf(Output, "%12.6f ", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[p]);
  }
  fprintf(Output, "\n");
  
  fprintf(Output, "\tNumber of variables    = %d\n", Simulation->NeuralModels[ModelNum].NumberOfVariables);
  fprintf(Output, "\tInitial values         = ");

  // Display of initial values
  for(int p = 0; p < Simulation->NeuralModels[ModelNum].NumberOfVariables; p++){
    fprintf(Output, "%12.6f ", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InitialValues[p]);
  }
  fprintf(Output, "\n");
  
  fprintf(Output, "\tPotential index        = %d\n", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].PotentialIndex);
  fprintf(Output, "\tNumber of spikes       = %d\n", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfSpikes);
  if(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].LastSpikeTime > -1.0e+300){
    fprintf(Output, "\tLast spike time        = %.5f\n", Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].LastSpikeTime);
  }
  fprintf(Output, "\n");
}

/*!*****************************************************************************
 * @brief Display of the execution times of different subparts of the program
 *
 * Display of execution times for the specified thread or all threads if value is negative.
 *
 * @param Simulation Structure where all simulation information is stored
 * @param Thread     Thread number (all threads are displayed for negative values)
 * @param Output     Output stream of the display
 ******************************************************************************/
void PrintTimes(Simul* Simulation, int Thread, FILE *Output)
{
  int Start = Thread, End = Thread+1;
    
  if(Thread < 0){
    // Settings for display of all threads
    Start = 0, End = Simulation->NbThreads;
  }
  
  // Display for threads between Start and End-1
  for(int i = Start; i < End && i < MAXTHREADS; ++i){
    fprintf(Output, "%2d Initialization time                             = %lf s\n", i, Simulation->Time_Init[i]);
    fprintf(Output, "%2d Advance time                                    = %lf s\n", i, Simulation->Time_Advance[i]);
    fprintf(Output, "%2d   Iteration time                                = %lf s\n", i, Simulation->Time_Iter[i]);
    fprintf(Output, "%2d     Numerical Method time                       = %lf s\n", i, Simulation->Time_NumericalMethod[i]);
    fprintf(Output, "%2d       UpdateModels time                         = %lf s\n", i, Simulation->Time_UpdateModels[i]);
    fprintf(Output, "%2d         UpdateNeuron time                       = %lf s\n", i, Simulation->Time_UpdateNeuron[i]);
    fprintf(Output, "%2d     Spike detection time                        = %lf s\n", i, Simulation->Time_Detection[i]);
    fprintf(Output, "%2d       Spike processing time                     = %lf s\n", i, Simulation->Time_SpikeProcessing[i]);
    fprintf(Output, "%2d         Random generation of connectivity time  = %lf s\n", i, Simulation->Time_RandGenConn[i]);
  }
}
