/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Internal library for generating pseudo-random numbers
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>

#include "GenLib.h"

/*!*****************************************************************************
 * @brief Initialization of the PRNG using Splitmix64 from S. Vigna and D. Lemire
 *
 * @param Seed  Initialization value of the PRNG
 * @param State Initialized state of the PRNG
 ******************************************************************************/
void InitPRNG(uint64_t Seed, PRNGState *State)
{
  State->State = (((__uint128_t) Splitmix64Stateless(Seed)) << 64) + Splitmix64Stateless(Seed + 1);
  State->Index = 0;
  State->Value.l = 0;
}

/*!*****************************************************************************
 * @brief Generation of one pseudo-random int at a time using Lehmer64 RNG (D. Lemire version)
 *
 * @see https://lemire.me/blog/2019/03/19/the-fastest-conventional-random-number-generator-that-can-pass-big-crush
 *
 * @param State PRNG state used to generate the next value
 *
 * @result A pseudo-random unsigned int
 ******************************************************************************/
uint32_t ValPRNGi(PRNGState *State)
{
  if(State->Index == 0){
    // Generation of a new 64 bits word only when needed
    State->State *= UINT64_C(0xda942042e4dd58b5);
    State->Value.l = State->State >> 64;
  }
  // Selection of the next generated byte
  uint32_t val = State->Value.i[State->Index];
  State->Index = (State->Index ^ 1) & 1; // & 1 is equivalent to mod 2
  return val;
}

/*!*****************************************************************************
 * @brief Generation of one pseudo-random long int at a time using Lehmer64 RNG (D. Lemire version)
 *
 * @see https://lemire.me/blog/2019/03/19/the-fastest-conventional-random-number-generator-that-can-pass-big-crush
 *
 * @param State PRNG state used to generate the next value
 *
 * @result A pseudo-random unsigned long int
 ******************************************************************************/
uint64_t ValPRNGl(PRNGState *State)
{
  // Generation of a new 64 bits word
  State->State *= UINT64_C(0xda942042e4dd58b5);
  State->Value.l = State->State >> 64;
  return State->Value.l;
}
