/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Internal library for generating pseudo-random numbers
 ******************************************************************************/

#ifndef GENLIB_H
#define GENLIB_H

#include <stdint.h>
#include <math.h>

/*!*****************************************************************************
 * @brief Macro for random real generator complient to multi-threading
 *
 * The generated number is between 0 and 1
 ******************************************************************************/
#define RandomReal(State) ((double)ValPRNGl(&State) / (~0UL))

/*!*****************************************************************************
 * @brief Macro for symmetric uniform random number generator complient to multi-threading
 *
 * The generated number is between -Bound and Bound
 ******************************************************************************/
#define UniformNoise(State, Bound) (Bound * (2.0 * RandomReal(State) - 1.0))

/*!*****************************************************************************
 * @brief Macro for specific noise function complient to multi-threading
 ******************************************************************************/
#define SpecificNoise(State, Bound) (1.0 + UniformNoise(State, Bound))

/*******************************************************************************
 * Unsigned byte integer
 ******************************************************************************/
typedef unsigned char uchar;

/*!*****************************************************************************
 * @brief Polymorphic representation of the PRNG value
 ******************************************************************************/
typedef union {
  uint64_t l;        ///< Interpretation as one long integer
  uint32_t i[2];     ///< Interpretation as two standard integers
} UnionPRNG;

/*!*****************************************************************************
 * @brief PRNG state
 ******************************************************************************/
typedef struct {
  __uint128_t State; ///< Internal state
  uchar       Index; ///< Current index of byte to be produced by the PRNG
  UnionPRNG   Value; ///< Produced value
} PRNGState;

/*!*****************************************************************************
 * @brief Splitmix64 function from S. Vigna and D. Lemire, without stored state
 *
 * Function used for the initialization of the PRNG
 *
 * @param Index Index used to generate first internal state of the PRNG
 ******************************************************************************/
static inline uint64_t Splitmix64Stateless(uint64_t Index) {
  uint64_t z = (Index + UINT64_C(0x9E3779B97F4A7C15));
  z = (z ^ (z >> 30)) * UINT64_C(0xBF58476D1CE4E5B9);
  z = (z ^ (z >> 27)) * UINT64_C(0x94D049BB133111EB);
  return z ^ (z >> 31);
}

/*!*****************************************************************************
 * @brief Initialization of the PRNG using Splitmix64 from S. Vigna and D. Lemire
 *
 * @param Seed  Initialization value of the PRNG
 * @param State Initialized state of the PRNG
 ******************************************************************************/
extern void InitPRNG(uint64_t Seed, PRNGState *State);

/*!*****************************************************************************
 * @brief Generation of one pseudo-random int at a time using Lehmer64 RNG (D. Lemire version)
 *
 * @see https://lemire.me/blog/2019/03/19/the-fastest-conventional-random-number-generator-that-can-pass-big-crush
 *
 * @param State PRNG state used to generate the next value
 *
 * @result A pseudo-random unsigned int
 ******************************************************************************/
extern uint32_t ValPRNGi(PRNGState *State);

/*!*****************************************************************************
 * @brief Generation of one pseudo-random long int at a time using Lehmer64 RNG (D. Lemire version)
 *
 * @see https://lemire.me/blog/2019/03/19/the-fastest-conventional-random-number-generator-that-can-pass-big-crush
 *
 * @param State PRNG state used to generate the next value
 *
 * @result A pseudo-random unsigned long int
 ******************************************************************************/
extern uint64_t ValPRNGl(PRNGState *State);

#endif
