/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Different input currents of the neurons
 ******************************************************************************/

#include <math.h>
#include <stdlib.h>

#include "Sirene.h"

/*!*****************************************************************************
 * @brief Constant input current
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param t          Current time t 
 *
 * @result Value of the input current of neuron NeuronNum in model ModelNum at time t
 ******************************************************************************/
double ConstantInputCurrent(Simul *Simulation, int ModelNum, int NeuronNum, double t)
{
  // The constant value is in InputParameters[0]
  return (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[0]);
}

/*!*****************************************************************************
 * @brief Constant input current with symmetric uniform noise
 * optionally multiplied by a specific amplitude noise
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param t          Current time t 
 *
 * @result Value of the input current of neuron NeuronNum in model ModelNum at time t
 ******************************************************************************/
double InputCurrentWithNoise(Simul *Simulation, int ModelNum, int NeuronNum, double t)
{
  // The bound of the symmetric uniform noise is in InputParameters[1]
  double sigma = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[1];

  return (// Constant input current (in InputParameters[0]) plus...
          Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[0]
          // ...the symmetric uniform noise between [-sigma : +sigma]
          + UniformNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, sigma)
          // ...optionnaly multiplied by a specific noise
          // * SpecificNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, 0.5) )
          );
}

/*!*****************************************************************************
 * @brief Stepwise input current
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param t          Current time t 
 *
 * @result Value of the input current of neuron NeuronNum in model ModelNum at time t
 ******************************************************************************/
double StepwiseInputCurrent(Simul *Simulation, int ModelNum, int NeuronNum, double t)
{
  // The step time is in InputParameters[2]
  if(t < Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[2]){
    // The value before the step time is in InputParameters[0]
    return (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[0]);
  }else{
    // The value after the step time is in InputParameters[1]
    return (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[1]);
  }
}

/*!*****************************************************************************
 * @brief Stepwise input current with a specific amplitude noise
 * and an optional symmetric uniform noise
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param t          Current time t 
 *
 * @result Value of the input current of neuron NeuronNum in model ModelNum at time t
 ******************************************************************************/
double StepwiseInputCurrentWithNoise(Simul *Simulation, int ModelNum, int NeuronNum, double t)
{
  // The amplitude of the specific noise is in InputParameters[3]
  double sigma = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[3];
 
  // The step time is in InputParameters[2]
  if(t < Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[2]){
    // Value before step time (InputParameters[0]) multiplied by a specific noise
    return (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[0] 
          * SpecificNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, sigma));
    // Optional addition of a symmetric uniform noise in [-sigma : +sigma]
    // + UniformNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, sigma)); 
  }else{
    // Value after step time (InputParameters[1]) multiplied by a specific noise
    return (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters[1]
          * SpecificNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, sigma));
    // Optional addition of a symmetric uniform noise in [-sigma : +sigma]
    // + UniformNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, sigma));
  }
}
