/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief MSN neural network 
 * 
 * McCarthy et al 2011
 * @see https://www.pnas.org/content/pnas/108/28/11620.full.pdf
 ******************************************************************************/

#include <stdio.h>
#include <math.h>

#include "Sirene.h"
#include "MSN.h"

/*******************************************************************************
 * Membrane capacitance (from McCarthy et al 2011)
 ******************************************************************************/
const double Cm = 1.0; ///< Membrane capacitance (in µF)

/*******************************************************************************
 * Leak current (from McCarthy et al 2011)
 ******************************************************************************/
const double EL = -67.0; ///< Ionic reversal Potential of leak current (in mV)

inline double I_L_MSN(double u) {return (u - EL); } ///< Leak current (in microA)

/*******************************************************************************
 * Na current (from McCarthy et al 2011)
 ******************************************************************************/
const double ENa = 50.0; ///< Ionic reversal Potential of Na current (in mV)

// Rate functions
inline double alpha_m(double u)   { double tmp = u + 54.0; return 0.32 * tmp / (1.0 - exp( - tmp / 4.0)); } ///< Rate function of Na current activation (m) variable (in 1/ms)
inline double beta_m(double u)    { double tmp = u + 27.0; return 0.28 * tmp / (exp( tmp / 5.0) - 1.0); }   ///< Rate function of Na current activation (m) variable (in 1/ms)

inline double alpha_h(double u)   { return 0.128 * exp( -(u + 50.0) / 18.0); }      ///< Rate function of Na current inactivation (h) variable (in 1/ms)
inline double beta_h(double u)    { return 4.0 / (1.0 + exp( -(u + 27.0) / 5.0)); } ///< Rate function of Na current inactivation (h) variable (in 1/ms)

// Steady-state functions
inline double m_inf_MSN(double u) { double tmp = alpha_m(u); return tmp / (tmp + beta_m(u)); } ///< Steady-state function of Na current activation (m) variable (in 1)
inline double h_inf_MSN(double u) { double tmp = alpha_h(u); return tmp / (tmp + beta_h(u)); } ///< Steady-state function of Na current inactivation (h) variable (in 1)

// Time constants of decay
inline double Tau_m_MSN(double u) { return 1.0 / (alpha_m(u) + beta_m(u)); } ///< Time constant of decay of Na current activation (m) variable (in ms)
inline double Tau_h_MSN(double u) { return 1.0 / (alpha_h(u) + beta_h(u)); } ///< Time constant of decay of Na current inactivation (h) variable (in ms)

// Na Current
inline double I_Na_MSN(double u, double m, double h) { return  m * m * m * h * (u - ENa); } ///< Na Current (in microA)

/*******************************************************************************
 * K current (from McCarthy et al 2011)
 ******************************************************************************/
const double EK = -100.0; ///< Ionic reversal Potential of K current (in mV)

// Rate functions
inline double alpha_n(double u)           { double tmp = u + 52.0; return 0.032 * tmp / (1.0 - exp( -tmp / 5.0)); } ///< Rate function of K current activation (n) variable (in 1/ms)
inline double beta_n(double u)            { return 0.5 * exp( -(u + 57.0) /40.0); }                                 ///< Rate function of K current activation (n) variable (in 1/ms)

// Steady-state functions
inline double n_inf_MSN(double u)         { double tmp = alpha_n(u); return tmp / (tmp + beta_n(u)); } ///< Steady-state function of K current activation (n) variable (in 1)

// Time constants of decay
inline double Tau_n_MSN(double u)         { return 1.0 / (alpha_n(u) + beta_n(u)); }                   ///< Time constant of decay of K current activation (n) variable (in ms)

// K Current
inline double I_K_MSN(double u, double n) { double tmp = n * n; return tmp * tmp * (u - EK); }         ///< K Current (in microA)

/*******************************************************************************
 * M current (from McCarthy et al 2011)
 ******************************************************************************/

// Rate functions
inline double alpha_p(double u)       { double tmp = u + 30.0; return 0.0003209 *  tmp / (1.0 - exp( -tmp / 9.0)); } ///< Rate function of M current activation (p) variable (in 1/ms)
inline double beta_p(double u)	      { double tmp = u + 30.0; return -0.0003209 * tmp / (1.0 - exp( tmp / 9.0)); }  ///< Rate function of M current activation (p) variable (in 1/ms)

// Steady-state functions
inline double p_inf(double u)         { double tmp = alpha_p(u); return tmp / (tmp + beta_p(u)); } ///< Steady-state function of M current activation (p) variable (in 1)

// Time constants decay
inline double Tau_p(double u)	      { return 1.0 / (alpha_p(u) + beta_p(u)); }                   ///< Time constant of decay of M current activation (p) variable (in ms)

// M Current
inline double I_M(double u, double p) { return p * (u - EK); }                                     ///< M Current (in microA)

/*!*****************************************************************************
 * @brief MSN neuron derivative equations 
 *
 * @param Simulation  Structure where all simulation information is stored
 * @param ModelNum    Neural model number
 * @param NeuronNum   Neuron number
 * @param t           Current time t
 * @param Variables   Gating variables of the neural model at the current time t
 * @param Derivatives Gating variables derivatives at the current time t
 ******************************************************************************/
void MSN(Simul *Simulation, int ModelNum, int NeuronNum, double t, double* Variables, double* Derivatives)
{	
  double u, m, n, h, p, gL, gNa, gK, gM; // Descriptions given below

  // Gating Variables
  u = Variables[0]; // Membrane potential (in mV)
  m = Variables[1]; // Activation gating variable of the Na current (in 1)
  h = Variables[2]; // Inactivation gating variable of the Na current (in 1)
  n = Variables[3]; // Activation gating variable of the K current (in 1)
  p = Variables[4]; // Activation gating variable of the M current (in 1)

  // Conductances of the different currents
  gL  = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[0]; // Leak conductance (in mS)
  gNa = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[1]; // Na current conductance (in mS)
  gK  = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[2]; // K current conductance (in mS)
  gM  = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[3]; // M current conductance (in mS)

  // Voltage derivative
  Derivatives[0] = (- gL * I_L_MSN(u) - gNa * I_Na_MSN(u, m, h) - gK * I_K_MSN(u, n) - gM * I_M(u, p)
                    - Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent + Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputCurrent
                   ) / Cm; // Membrane potential (in mV/ms)
  
  // Gating variables derivatives
  Derivatives[1] = (m_inf_MSN(u) - m) / Tau_m_MSN(u); // Activation gate of the Na current (in 1/ms)
  Derivatives[2] = (h_inf_MSN(u) - h) / Tau_h_MSN(u); // Inactivation gate of the Na current (in 1/ms)
  Derivatives[3] = (n_inf_MSN(u) - n) / Tau_n_MSN(u); // Activation gate of the K current (in 1/ms)
  Derivatives[4] = (p_inf(u) - p) / Tau_p(u);         // Activation gate of the M current (in 1/ms)
}
