/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Memory management functions
 ******************************************************************************/

#include "Memory.h"

/*!*****************************************************************************
 * @brief Memory allocation of synapse models
 *
 * @param NbModels Numbers of neural models
 *
 * @result A pointer to the newly allocated 2D array
 ******************************************************************************/
SynapseModel **SynapseModelsAlloc(int NbModels)
{
  SynapseModel **tab;

  tab = (SynapseModel **) MYMALLOC(NbModels * sizeof(SynapseModel *));
  tab[0] = (SynapseModel *) MYMALLOC(NbModels * NbModels * sizeof(SynapseModel));
  for(int i = 1; i < NbModels; ++i){
    tab[i] = tab[i-1] + NbModels;
  }
  return tab;
}

/*!*****************************************************************************
 * @brief Memory de-allocation of all dynamic variables used in the simulation
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
void MemoryRelease(Simul *Simulation)
{
  // For every neural model...
  for(int ModelNum = 0; ModelNum < Simulation->NumberNeuralModels; ModelNum++){
    // Deallocation of every neuron
    for(int NeuronNum = 0; NeuronNum < Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
      MYFREE(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InitialValues);
      MYFREE(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputParameters);
      MYFREE(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters);
      MYFREE(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[0]);
      MYFREE(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor);
    }
    // Deallocation of all parameters
    for(int DstModel = 0; DstModel < Simulation->NumberNeuralModels; DstModel++){
      MYFREE(Simulation->SynapseModels[ModelNum][DstModel].Parameters);
    }
    MYFREE(Simulation->NeuralModels[ModelNum].Parameters);
    MYFREE(Simulation->NeuralModels[ModelNum].Neurons);
  }
  // Deallocation of data arrays in the simulation
  MYFREE(Simulation->NeuralModels);
  MYFREE(Simulation->SynapseModels[0]);
  MYFREE(Simulation->SynapseModels);
  MYFREE(Simulation->Variables[0]);
  MYFREE(Simulation->DVar[0]);
  MYFREE(Simulation->BufferSpikes);
}
