/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Memory management functions
 ******************************************************************************/

#ifndef MEMORY_H
#define MEMORY_H

#include "Sirene.h"

/*!*****************************************************************************
 * @brief Memory allocation of synapse models
 *
 * @param NbModels Numbers of neural models
 *
 * @result A pointer to the newly allocated 2D array
 ******************************************************************************/
extern SynapseModel **SynapseModelsAlloc(int NbModels);

/*!*****************************************************************************
 * @brief Memory de-allocation of all dynamic variables used in the simulation
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
extern void MemoryRelease(Simul *);

/*******************************************************************************
 * Memory management data
 ******************************************************************************/
extern unsigned long MEMORY;    ///< Total amount of requested memory
extern unsigned long ALLOCATED; ///< Total amount of actually allocated memory
extern unsigned long FREED;     ///< Total amount of de-allocated memory

/*!*****************************************************************************
 * @brief Inline allocation function with profiling of memory consumption
 *
 * @param Size Number of bytes to allocate
 *
 * @result A pointer to the newly allocated memory
 ******************************************************************************/
inline void *MYMALLOC(size_t Size)
{
  void *ret = malloc(Size);
  MEMORY += Size;
  ALLOCATED += malloc_usable_size(ret);
  return ret;
}

/*!*****************************************************************************
 * @brief Inline free function with profiling of memory consumption
 *
 * @param Ptr Pointer to deallocate
 ******************************************************************************/
inline void MYFREE(void *Ptr)
{
  if(Ptr != NULL){
    long Size = malloc_usable_size(Ptr);
    if(Size > 0){
      FREED += Size;
      free(Ptr);
    }else{
      printf("%p pointer to be released with invalid size : %ld\n", Ptr, Size);
    }
  }
}

/*!*****************************************************************************
 * @brief Memory consumption report
 ******************************************************************************/
inline void DISPLAYMEM()
{
  printf("Memory check : %lu (A: %lu F: %lu Diff: %ld)\n", MEMORY, ALLOCATED, FREED, (long)ALLOCATED-FREED);
}

#endif
