/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Numerical methods used to compute the network evolution at each time-step
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "Sirene.h"

/*!*****************************************************************************
 * @brief Resolution of the ODE system by explicit Euler method
 *
 * @param Simulation Structure where all simulation information is stored
 * @param Update     Update function of each neuron in the models
 * @param TmpVar     Not used
 * @param TmpDer     Not used
 ******************************************************************************/
void Euler(Simul  *Simulation,
           void (*Update)(Simul *Simulation, double tPrev, double t, double *Var, double *Der),
           double *TmpVar, double *TmpDer)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_NumericalMethod);

  // Update of each neuron in each model for the first simulation time step
  if((Simulation->Start <= Simulation->t[Simulation->PrvNum])  &&  (Simulation->t[Simulation->PrvNum] < Simulation->TimeStep)){
    (*Update)(Simulation, Simulation->t[Simulation->PrvNum] - Simulation->TimeStep, Simulation->t[Simulation->PrvNum], Simulation->Variables[Simulation->PrvNum], Simulation->DVar[Simulation->PrvNum]);
  }

  // Computation of the Variables at next time step (t + dt)
  #pragma omp for
  for(int i = 0; i < Simulation->NumberOfVariables; i++){
    Simulation->Variables[Simulation->CrtNum][i] = Simulation->Variables[Simulation->PrvNum][i] + Simulation->DVar[Simulation->PrvNum][i] * Simulation->TimeStep;
  }

  // Update of each neuron in each model at the time t + dt
  (*Update)(Simulation, Simulation->t[Simulation->PrvNum], Simulation->t[Simulation->CrtNum], Simulation->Variables[Simulation->CrtNum], Simulation->DVar[Simulation->CrtNum]);

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_NumericalMethod, Simulation->Time_NumericalMethod);
}

/*!*****************************************************************************
 * @brief Resolution of the ODE system by the explicit Runge-Kutta 2 method
 *
 * @param Simulation Structure where all simulation information is stored
 * @param Update     Update function of each neuron in the models
 * @param TmpVar     Variables array for intermediate time-steps
 * @param TmpDer     Derivatives array for intermediate time-steps
 ******************************************************************************/
void RungeKutta2( Simul  *Simulation,
                  void (*Update)(Simul *Simulation, double tPrev, double t, double *Var, double *Der),
                  double *TmpVar, double *TmpDer)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_NumericalMethod);

  double HalfTimeStep = 0.5 * Simulation->TimeStep;

  // Update of each neuron in each model for the first simulation time step
  if((Simulation->Start <= Simulation->t[Simulation->PrvNum])  &&  (Simulation->t[Simulation->PrvNum] < Simulation->TimeStep)){
    (*Update)(Simulation, Simulation->t[Simulation->PrvNum], Simulation->t[Simulation->PrvNum], Simulation->Variables[Simulation->PrvNum], Simulation->DVar[Simulation->PrvNum]);
  }
  
  // Computation of the variables at time the t + 0.5 * dt
  #pragma omp for
  for(int i = 0; i < Simulation->NumberOfVariables; i++){
    TmpVar[i] = Simulation->Variables[Simulation->PrvNum][i] + HalfTimeStep * TmpDer[i];
  }

  // Update of each neuron in each model at the time t + dt/2
  (*Update)(Simulation, Simulation->t[Simulation->PrvNum], Simulation->t[Simulation->PrvNum] + HalfTimeStep, TmpVar, TmpDer); 

  // Computation of the variables at the time t + dt
  #pragma omp for
  for(int i=0; i<Simulation->NumberOfVariables; i++){
    Simulation->Variables[Simulation->CrtNum][i] = Simulation->Variables[Simulation->PrvNum][i] + Simulation->TimeStep * TmpDer[i];
  }

  // Update of each neuron in each model at time t + dt
  (*Update)(Simulation, Simulation->t[Simulation->PrvNum] + HalfTimeStep, Simulation->t[Simulation->CrtNum], Simulation->Variables[Simulation->CrtNum], Simulation->DVar[Simulation->CrtNum]);

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_NumericalMethod, Simulation->Time_NumericalMethod);
}
