/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Pseudo-random connectivity generator and post-synaptic neurons update
 ******************************************************************************/

#include <math.h>
#include <stdlib.h>

#include "Sirene.h"
#include "SynapticCurrent.h"

/*!*****************************************************************************
 * @brief Generation of pseudo-random post-synaptic connections of neuron NeuronNum in model ModelNum
 *
 * The SynapticSpike function is used to update the synaptic current of the post-synaptic neurons.
 *  
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param SpikeTime  Spike time
 ******************************************************************************/
void RandomConnections(Simul *Simulation, int ModelNum, int NeuronNum, double SpikeTime)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_RandGenConn);

  // Initialization of the internal state of the random generator associated to the neuron
  PRNGState State;
  uint64_t Seed = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron;
  InitPRNG(Seed, &State);

  // Run through post-synaptic neural models
  for(int PostSynModel = 0; PostSynModel < Simulation->NumberNeuralModels; PostSynModel++){
    
    // Generation is done only when there are synapses from model ModelNum to model PostSynModel
    if(Simulation->SynapseModels[ModelNum][PostSynModel].NumberOfParameters >= 0){
      // Number of selected postsynaptic neurons 
      uint NbSelected = 0;
      // Total number of neurons in neural model PostSynModel
      uint NbNeuronsPostSynModel = Simulation->NeuralModels[PostSynModel].NumberOfNeurons;
      // Number of connections between model ModelNum and model PostSynModel
      uint NbConnections = Simulation->SynapseModels[ModelNum][PostSynModel].NumberOfConnections;

      // Post-synaptic model is different from the current model
      if(ModelNum != PostSynModel){
        if(NbConnections == NbNeuronsPostSynModel){
          // Simple case of 100% connectivity
          for(uint PostNeuronNum = 0; PostNeuronNum < NbNeuronsPostSynModel; ++PostNeuronNum) {
            SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostNeuronNum, SpikeTime);
          }
        }else{
          // Post-synaptic neurons generation and synaptic currents update 
          for(uint PostNeuronNum = 0; PostNeuronNum < NbNeuronsPostSynModel && NbSelected < NbConnections; ++PostNeuronNum) {
            uint RemainingNeurons = NbNeuronsPostSynModel - PostNeuronNum; 
            uint RemainingConnections = NbConnections - NbSelected;
            if(ValPRNGi(&State) % RemainingNeurons < RemainingConnections) {
              int PostSynNeuron = PostNeuronNum; 
              SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostSynNeuron, SpikeTime);
              NbSelected++;
            }
          }
        }

      } else { // Post-synaptic model is the current model

        if(NbConnections == NbNeuronsPostSynModel){
          // Simple case of 100% connectivity with removal of the self-connection
          for(uint PostNeuronNum = 0; PostNeuronNum < ((uint)NeuronNum); ++PostNeuronNum) {
            SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostNeuronNum, SpikeTime);
          }
          for(uint PostNeuronNum = NeuronNum+1; PostNeuronNum < NbNeuronsPostSynModel; ++PostNeuronNum) {
            SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostNeuronNum, SpikeTime);
          }
        }else{
          // Generation and synaptic currents update of post-synaptic neurons before neuron NeuronNum
          for(uint PostNeuronNum = 0; PostNeuronNum < ((uint)NeuronNum) && PostNeuronNum < NbNeuronsPostSynModel && NbSelected < NbConnections; ++PostNeuronNum) {
            uint RemainingNeurons = NbNeuronsPostSynModel - PostNeuronNum - 1; // -1 to deduct the self-connection
            uint RemainingConnections = NbConnections - NbSelected;
            if(ValPRNGi(&State) % RemainingNeurons < RemainingConnections) {
              int PostSynNeuron = PostNeuronNum; 
              SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostSynNeuron, SpikeTime);
              NbSelected++;
            }
          }
          // Generation and synaptic currents update of post-synaptic neurons after neuron NeuronNum
          for(uint PostNeuronNum = NeuronNum+1; PostNeuronNum < NbNeuronsPostSynModel && NbSelected < NbConnections; ++PostNeuronNum) {
            uint RemainingNeurons = NbNeuronsPostSynModel - PostNeuronNum;
            uint RemainingConnections = NbConnections - NbSelected;
            if(ValPRNGi(&State) % RemainingNeurons < RemainingConnections) {
              int PostSynNeuron = PostNeuronNum;
              SynapticSpike(Simulation, ModelNum, NeuronNum, PostSynModel, PostSynNeuron, SpikeTime);
              NbSelected++;
            }
          }
        }
      }
    } // End if synaptic connections
  } // End for models
  
  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_RandGenConn, Simulation->Time_RandGenConn);
}
