/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Main functions of the SiReNe Software.
 *
 * Main functions of the iterative process to simulate biological neuronal networks
 * with adjustable properties (normal state, Parkinson's disease,...).
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "SireneComponents.h"

/*******************************************************************************
 * Declaration and initialization of memory management data
 ******************************************************************************/
unsigned long MEMORY=0, ALLOCATED=0, FREED=0;

/*!*****************************************************************************
 * @brief Computation of one time-step from (t) to (t + dt)
 *
 * Network state is computed at (t + dt) using a numerical method (Euler or Runge-Kutta 2).\n
 * Detection of spikes using the SpikeDetection() function.
 *
 * @param Simulation Structure where all simulation information is stored
 * @param TmpVar     Variables array for intermediate time-steps computed in the numerical method
 * @param TmpDer     Derivatives array for intermediate time-steps computed in the numerical method
 ******************************************************************************/
void Iter(Simul *Simulation, double *TmpVar, double *TmpDer)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_Iter);
  
  //
  // Update of the time variable from (t) to (t+dt): only one thread
  //
  #pragma omp single
  {
    Simulation->t[Simulation->CrtNum] = Simulation->t[Simulation->PrvNum] + Simulation->TimeStep;
  }

  //
  // Resolution of the ODE system by approximation with a numerical method: all threads
  //
  (*Simulation->NumericalMethod)(Simulation, UpdateModels, TmpVar, TmpDer);  

  //
  // Synchronization of all threads before applying the spikes
  //
  #pragma omp barrier

  //
  // Spikes detection and synaptic current updating for the post-synaptic neurons: all threads
  //
  SpikeDetection(Simulation);

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_Iter, Simulation->Time_Iter);
}

/*!*****************************************************************************
 * @brief Simulation process over time-steps
 *
 * Iterative process over the time-steps:
 *   - Calls to the Iter() function to solve the ODE system of the network at each time-step
 * 
 * Possibility to save data like VARIABLES, POTENTIALS, INPUT_CURRENTS, SYNAPTIC_CURRENTS, LFP, SPIKES and NBSPIKES.\n
 * All files are saved in binary mode.
 *
 * @param Simulation   Structure where all simulation information is stored
 ******************************************************************************/
void Advance(Simul *Simulation)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_Advance);

  int sic;                        // Save interval counter
  int dtc;                        // Display time counter
  int tmpt;                       // Temporary time variable
  double Start = omp_get_wtime(); // Start time
  double Duration;                // Execution time
  double TimeStep;                // Current time-step
  double *TmpVar = NULL;          // Array of intermediate variables
  double *TmpDer = NULL;          // Array of intermediate derivatives

  //
  // Intermediate arrays allocation
  //
  TmpDer = (double *) MYMALLOC(Simulation->NumberOfVariables*sizeof(double));
  TmpVar = (double *) MYMALLOC(Simulation->NumberOfVariables*sizeof(double));
  
  //
  // Simulation with data saving
  //
  if(Simulation->SaveMask != NO_RECORD){
    FILE *VariablesFile = NULL, *PotentialFile = NULL, *InputCurrentFile = NULL;
    FILE *SynapticCurrentFile = NULL, *LFPFile = NULL, *NbSpikesFile = NULL;

    // The first line of the file is the global number of variables or total number of neurons during the simulation
    // -> It may be useful to turn the binary file into a text file with a python script
    if((Simulation->SaveMask & SAVE_VARIABLES) != 0){
      VariablesFile = fopen(Simulation->VariablesFileName, "w");
      fwrite(&Simulation->NumberOfVariables, sizeof(int), 1, VariablesFile); 
    }
    if((Simulation->SaveMask & SAVE_POTENTIALS) != 0){
      PotentialFile = fopen(Simulation->PotentialsFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, PotentialFile);
    }
    if((Simulation->SaveMask & SAVE_INPUT_CURRENTS) != 0){
      InputCurrentFile = fopen(Simulation->InputCurrentsFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, InputCurrentFile);
    }
    if((Simulation->SaveMask & SAVE_SYNAPTIC_CURRENTS) != 0){
      SynapticCurrentFile = fopen(Simulation->SynapticCurrentsFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, SynapticCurrentFile);
    }
    if((Simulation->SaveMask & SAVE_LFP) != 0){
      LFPFile = fopen(Simulation->LFPFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, LFPFile);
    }
    if((Simulation->SaveMask & SAVE_SPIKES) != 0){
      Simulation->SpikesFile = fopen(Simulation->SpikesFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, Simulation->SpikesFile);
    }
    if((Simulation->SaveMask & SAVE_NBSPIKES) != 0){
      NbSpikesFile = fopen(Simulation->NbSpikesFileName, "w");
      fwrite(&Simulation->TotalNumberOfNeurons, sizeof(int), 1, NbSpikesFile);
    }
    
    //
    // Multithreaded part
    //
    #pragma omp parallel num_threads(Simulation->NbThreads) private(TimeStep) if(Simulation->NbThreads > 1)
    {
      //
      // Init of sic counter (save interval counter) and dtc counter (display time counter): only one thread
      //
      #pragma omp single
      {sic = 0; dtc = 0;}
      
      //
      // MAIN LOOP: all threads
      //
      for(TimeStep=Simulation->Start; TimeStep<=Simulation->TotalTimeSim; TimeStep+=Simulation->TimeStep){

        //
        // Buffers exchange and time step management: only one thread
        //
        #pragma omp single
        {
          // Exchange of buffers at times t and t+dt
          tmpt               = Simulation->PrvNum;
          Simulation->PrvNum = Simulation->CrtNum;
          Simulation->CrtNum = tmpt;
        }

        //
        // Current simulation iteration: all threads
        //
        Iter(Simulation, TmpVar, TmpDer);

        //
        // Data saved in files: only one thread
        //
        #pragma omp single
        {
          sic++;
          dtc++;
          if(sic > Simulation->ResultsInterval){
            // Simulation duration display
            if(dtc > 1e5){
              Duration   = omp_get_wtime() - Start;
              fprintf(stderr, "INFO: %.3lf%% du calcul en %.2lfs\n", 100 * TimeStep / Simulation->TotalTimeSim, Duration);
              dtc = 0;
            }

            sic = 0;

            // Variables saving
            if((Simulation->SaveMask & SAVE_VARIABLES) !=0){
              fwrite(&Simulation->t[Simulation->PrvNum], sizeof(double), 1, VariablesFile);
              for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
                for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
                  for(int i = 0;i < Simulation->NumberOfVariables;i++){
                    fwrite(&Simulation->Variables[Simulation->PrvNum][i], sizeof(double), 1, VariablesFile);
                  }
                }
              }
              fflush(VariablesFile);
            }
        
            // Potentials saving
            if((Simulation->SaveMask & SAVE_POTENTIALS) !=0){
              fwrite(&Simulation->t[Simulation->PrvNum], sizeof(double), 1, PotentialFile);
              for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
                for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
                  fwrite(&Simulation->Variables[Simulation->PrvNum][Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].PotentialIndex], sizeof(double), 1, PotentialFile);
                }
              }
              fflush(PotentialFile);
            }
        
            // Inputs currents saving
            if((Simulation->SaveMask & SAVE_INPUT_CURRENTS) != 0){
              fwrite(&Simulation->t[Simulation->PrvNum], sizeof(double), 1, InputCurrentFile);
              for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
                for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
                  fwrite(&Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputCurrent, sizeof(double), 1, InputCurrentFile);
                }
              }
              fflush(InputCurrentFile);
            }

            // Synaptics currents saving
            if((Simulation->SaveMask & SAVE_SYNAPTIC_CURRENTS) != 0){
              fwrite(&Simulation->t[Simulation->PrvNum], sizeof(double), 1, SynapticCurrentFile);
              for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
                for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
                  fwrite(&Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent, sizeof(double), 1, SynapticCurrentFile);
                }
              }
              fflush(SynapticCurrentFile);
            }

            // LFP saving
            if((Simulation->SaveMask & SAVE_LFP) != 0){
              Simulation->sLFP = 0;
              fwrite(&Simulation->t[Simulation->PrvNum], sizeof(double), 1, LFPFile);
              for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
                for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
                  Simulation->sLFP += Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent;
                }
              }

              if(Simulation->TotalNumberOfNeurons != 0){
                Simulation->sLFP = Simulation->sLFP/Simulation->TotalNumberOfNeurons;
                fwrite(&Simulation->sLFP, sizeof(double), 1, LFPFile);
              }else{
                fprintf(stderr, "Problem LFP saving: Division by 0 \n");
                exit(EXIT_FAILURE);
              }
              fflush(LFPFile);
            }
          } // End ResultsInterval
        } // End single section
      } // End for TimeStep
    } // End parallel section

    //
    // Spikes saving
    //
    if((Simulation->SaveMask & SAVE_SPIKES) != 0){
      // Save the remaining spikes
      if(Simulation->NbSpInBuf > 0){
        fwrite(Simulation->BufferSpikes, sizeof(SpikeEvent), Simulation->NbSpInBuf, Simulation->SpikesFile);
        fflush(Simulation->SpikesFile);
      }
    }

    //
    // Spikes frequencies saving of all neurons
    // 
    if((Simulation->SaveMask & SAVE_NBSPIKES) != 0){
      for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
        for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
          fwrite(&Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron, sizeof(uint), 1, NbSpikesFile);
          fwrite(&Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfSpikes, sizeof(int), 1, NbSpikesFile);
        }
      }
    }

    //
    // Files closing
    //
    if((Simulation->SaveMask & SAVE_VARIABLES)         != 0){ fclose(VariablesFile); }
    if((Simulation->SaveMask & SAVE_POTENTIALS)        != 0){ fclose(PotentialFile); }
    if((Simulation->SaveMask & SAVE_INPUT_CURRENTS)    != 0){ fclose(InputCurrentFile); }
    if((Simulation->SaveMask & SAVE_SYNAPTIC_CURRENTS) != 0){ fclose(SynapticCurrentFile); }
    if((Simulation->SaveMask & SAVE_LFP)               != 0){ fclose(LFPFile); }
    if((Simulation->SaveMask & SAVE_SPIKES)            != 0){ fclose(Simulation->SpikesFile); }
    if((Simulation->SaveMask & SAVE_NBSPIKES)          != 0){ fclose(NbSpikesFile); }

         //
  }else{ // Version without data saving
         //

    //
    // Multithreaded part
    //
    #pragma omp parallel num_threads(Simulation->NbThreads) private(TimeStep) if(Simulation->NbThreads > 1)
    {
      //
      // Init of dtc counter (display time counter): only one thread
      //
      #pragma omp single
      {dtc = 0;}
  
      //
      // MAIN LOOP: all threads
      //
      for(TimeStep=Simulation->Start; TimeStep<=Simulation->TotalTimeSim; TimeStep+=Simulation->TimeStep){

        //
        // Buffers exchange and time step management: only one thread
        //
        #pragma omp single
        {
          // Exchange of buffers at times t and t+dt
          tmpt                = Simulation->PrvNum;
          Simulation->PrvNum = Simulation->CrtNum;
          Simulation->CrtNum = tmpt;
  
          // Periodic display
          if(dtc == Simulation->ResultsInterval){
            dtc = 0;
            Duration   = omp_get_wtime() - Start;
            printf("INFO: %.3lf%% of the calculation in %.2lfs\n", 100 * TimeStep / Simulation->TotalTimeSim, Duration);
          }else{
            dtc++;
          }
        }

        //
        // Current simulation iteration: all threads
        //
        Iter(Simulation, TmpVar, TmpVar);

      } // End for TimeStep
    } // End parallel section
  } // End else over Simulation->SaveMask
  
  // Simulated time (final time) display
  printf("Final simulation time : %lf s\n", Simulation->t[Simulation->PrvNum]);
  printf("Last simulated time   : %lf s\n", Simulation->t[Simulation->CrtNum]);
  
  // Recovery and display of the real time, spent in the simulation
  Duration = omp_get_wtime() - Start;
  printf("TOTAL SIMULATION TIME : %lfs\n", Duration);

  //
  // Deallocation or intermediate arrays
  //
  MYFREE(TmpDer);
  MYFREE(TmpVar);

  // Post-processing for execution time profiling
  // -> Must be called BEFORE PrintTimes() function
  RETURN_CALL(Simulation->Chrono_Advance, Simulation->Time_Advance);

  //
  // Detailed display of execution times
  //
  #ifdef TIME_CALLS
  PrintTimes(Simulation, 0, stdout);
  #endif
}

/*!*****************************************************************************
 * @brief Initialization of the simulation structure
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
void InitSimu(Simul *Simulation)
{
  // Verbose
  Simulation->VerboseMask = 0;               // No verbose

  // Models
  Simulation->NeuralModels         = NULL;   // Empty list of models
  Simulation->SynapseModels        = NULL;   // Empty list of synaptic models
  Simulation->NumberNeuralModels   = 0;      // No neural model
  Simulation->NumberSynapseModels  = 0;      // No synapse
  Simulation->TotalNumberOfNeurons = 0;      // No neuron
  Simulation->NumberOfVariables    = 0;      // No variable
  Simulation->sLFP                 = 0;      // Empty sum

  // Computation types
  Simulation->InterpolationType = NO_INTERP; // Simplest interpolation

  // Simulated time management
  Simulation->Start        = 0;              // Simulation start
  Simulation->TimeStep     = 5e-3;           // Time step (maximal value)
  Simulation->TotalTimeSim = 1000;           // Simulation execution time (1s)
  for(int i=0; i<NBMEMSTEP; ++i){
    Simulation->t[i]         = 0;
    Simulation->Variables[i] = NULL;
    Simulation->DVar[i]      = NULL;
  }
  Simulation->PrvNum = 0;                    // Index of previous time
  Simulation->CrtNum = 1;                    // Index of current time

  // Data saving management
  Simulation->SaveMask        = 0;           // No variable to save
  Simulation->ResultsInterval = (1<<31);     // Very large interval (no display / saving)
  
  // Generic functions
  Simulation->RandGenerConnect = RandomConnections; // Internal pseudo-random connectivity generation
  Simulation->NumericalMethod  = Euler;             // Simplest method

  // Spikes management
  Simulation->SpikesFile   = NULL;           // No spikes file
  Simulation->BufferSpikes = NULL;           // No spikes buffer
  Simulation->NbSpInBuf    = 0;              // No spikes stored in the buffer
  
  // Multi-threading management
  Simulation->NbThreads = 1;                 // Sequential version
}

/*!*****************************************************************************
 * @brief Preparation of simulation data according to the neural network description
 *
 * Initialization of:
 *   - Simulation structure:
 *     - SaveMask for data saving during the simulation
 *     - Time index PrvNum (previous time-step)
 *     - Crt (current time-step)
 *     - Time t0 = simulation starting point 
 *     - Total number of neurons
 *     - Number of variables 
 *     - Allocation of the variables array
 *     - Allocation of the derivatives of the variables array
 *     - Sum of the LFP
 *   - Neuron structure:
 *     - Input/Synaptic current
 *     - Global number neuron
 *     - Number of parameters
 *     - Parameters
 *     - Synaptic current factor
 *     - Spike threshold
 *     - Potential index 
 *     - Mid-time step potential 
 *     - Mid-time step derivative of the potential
 *     - Number of spikes
 *     - Last spike time
 *   - Message:
 *     - Display of initialization information
 *       - Initialization display is activated via the VerboseMask parameter
 *   - Chrono:
 *     - Time initalization for each function
 *
 * @param Simulation      Structure where all simulation information is stored
 * @param VerboseMask     Mask of information to display
 * @param SaveMask        Mask of variables to save to file
 * @param ResultsInterval Number of iterations between results display / save
 ******************************************************************************/
void PrepareSimu(Simul *Simulation, ushort VerboseMask, ushort SaveMask, int ResultsInterval) 
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_Init);

  //
  // Simulation parameters initialization
  //

  // Storage of the verbose mask in the simulation structure
  Simulation->VerboseMask = VerboseMask;

  // Storage of the save mask in the simulation structure
  Simulation->SaveMask = SaveMask;

  // Storage of the save mask in the simulation structure
  Simulation->ResultsInterval = ResultsInterval;

  // Time Indexes in the network states buffer
  // -> Only NBMEMSTEP successive time steps are stored in memory
  // -> By default, only two time steps are stored, the previous time and the current time
  Simulation->PrvNum = 0; // Previous time index
  Simulation->CrtNum = 1; // Current time index
  
  // Initialization of simulation start time
  Simulation->t[Simulation->CrtNum] = Simulation->Start;

  // Global and per model numbers of variables and numbers of neurons
  Simulation->TotalNumberOfNeurons = 0;
  Simulation->NumberOfVariables    = 0;
  for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
    Simulation->TotalNumberOfNeurons += Simulation->NeuralModels[ModelNum].NumberOfNeurons;
    Simulation->NumberOfVariables    += Simulation->NeuralModels[ModelNum].NumberOfVariables * Simulation->NeuralModels[ModelNum].NumberOfNeurons;
  }

  // Variables array construction that contains variables of all neurons in all groups for times t and (t + dt)
  if((VerboseMask & VERBOSE_CONFIG) != 0){
    printf("Allocating %d bytes for Variables...\n", (int)(NBMEMSTEP * Simulation->NumberOfVariables * sizeof(double)));
  }
  Simulation->Variables[0] = (double *) MYMALLOC(NBMEMSTEP * Simulation->NumberOfVariables * sizeof(double));
  for(int i=1; i<NBMEMSTEP; ++i){
    Simulation->Variables[i] = Simulation->Variables[i-1] + Simulation->NumberOfVariables;
  }

  // Derivatives array construction that contains variables of all neurons in all groups for times t and (t + dt)
  Simulation->DVar[0] = (double *) MYMALLOC(NBMEMSTEP * Simulation->NumberOfVariables * sizeof(double));
  for(int i=1; i<NBMEMSTEP; ++i){
    Simulation->DVar[i] = Simulation->DVar[i-1] + Simulation->NumberOfVariables;
  }

  // Initialization of the LFP sum (to save in a file)
  Simulation->sLFP = 0;

  //
  // Neurons
  //
  int v = 0;  // Potential index for the variable array
  for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
    for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
      
      // Input/synaptic current Initialization
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputCurrent    = 0.0;
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent = 0.0;
      
      // Global neuron number
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron = Simulation->NeuralModels[ModelNum].FirstNeuronNumber + NeuronNum;

      // Parameters construction and attribution to each neuron 
      // -> This is useful if the number of parameters changes between the neurons of the same model (for ex. with a noise)
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfParameters = Simulation->NeuralModels[ModelNum].NumberOfParameters;
      if(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfParameters == 0){ 
        Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters = NULL;
      }else{
        Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters = (double *) MYMALLOC(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfParameters * sizeof(double));
      }
      for(int i = 0; i < Simulation->NeuralModels[ModelNum].NumberOfParameters; i++){
        Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].Parameters[i] = Simulation->NeuralModels[ModelNum].Parameters[i];
      }

      // SynCurrentFactor construction
      // -> Intialization of the SynCurrentFactor useful to update the synaptic current when a spike is detected
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor    = (double **) MYMALLOC(Simulation->NumberNeuralModels * sizeof(double *)); 
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[0] = (double *) MYMALLOC(Simulation->NumberNeuralModels * Simulation->NbThreads * sizeof(double)); 
      memset(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[0], 0, Simulation->NumberNeuralModels * Simulation->NbThreads * sizeof(double));
      for(int i = 1; i < Simulation->NumberNeuralModels; i++){
        Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[i] = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[i-1] + Simulation->NbThreads;
      }

      // SpikeThreshold for each neuron
      // -> This is useful if there exists a fluctuation between the spike thresholds of the neurons of a same model
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SpikeTreshold = Simulation->NeuralModels[ModelNum].SpikeTreshold;

      // Potential index in the array Variables 
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].PotentialIndex = v;
      for(int i = 0; i < Simulation->NeuralModels[ModelNum].NumberOfVariables; i++){
        Simulation->Variables[Simulation->CrtNum][v+i] = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InitialValues[i];
      }

      // Initialization of variable V at mid-time (t+0.5dt) and (t+1.5dt) for each neuron
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].VMidStep[Simulation->PrvNum] = -1.00000e+308;
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].VMidStep[Simulation->CrtNum] = -1.00000e+308;

      // Initialization of variable dV/dt at time (t+0.5dt) and (t+1.5dt) for each neuron
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].dVMidStep[Simulation->PrvNum] = -1.00000e+308;
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].dVMidStep[Simulation->CrtNum] = -1.00000e+308;

      // Increased potential index
      v += Simulation->NeuralModels[ModelNum].NumberOfVariables;

      // Initialization of the number of spikes for each neuron
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfSpikes = 0;

      // Initialization of the last spike time
      Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].LastSpikeTime = -1.00000e+308;
    }
  }

  // Spikes buffer construction
  Simulation->BufferSpikes = (SpikeEvent *) MYMALLOC(BUFFER_SIZE * sizeof(SpikeEvent));
  Simulation->NbSpInBuf = 0;
  
  //
  // Timers initializaton
  //
  Simulation->ChronoStart = omp_get_wtime();
  for(int i=0; i<Simulation->NbThreads && i<MAXTHREADS; ++i){
    Simulation->Time_UpdateModels[i]    = 0;
    Simulation->Time_UpdateNeuron[i]    = 0;
    Simulation->Time_NumericalMethod[i] = 0;
    Simulation->Time_SpikeProcessing[i] = 0;
    Simulation->Time_Iter[i]            = 0;
    Simulation->Time_Advance[i]         = 0;
    Simulation->Time_Init[i]            = 0;
    Simulation->Time_Detection[i]       = 0;
    Simulation->Time_RandGenConn[i]     = 0;
  }

  //
  // Display of simulation information
  //
  if((VerboseMask & VERBOSE_CONFIG) != 0){
    printf("**** EVENT DRIVEN UPDATES ****\n");
    #ifdef _OPENMP
    printf("Parallel version with %d threads\n", Simulation->NbThreads);
    #else
    printf("Sequential version\n");
    #endif
  }
  
  // Detailed display of the neural network structure (models, neurons, synapses)
  if(VerboseMask & VERBOSE_DETAILS){
    for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
      NeuralModelPrint(Simulation, ModelNum, stdout);
      for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
        NeuronPrint(Simulation, ModelNum, NeuronNum, stdout);
      }
      for(int DstModel=0; DstModel<Simulation->NumberNeuralModels; DstModel++){
        SynapseModelPrint(Simulation, ModelNum, DstModel, stdout);
      }
    }
  }

  // Gobal parameters of the simulation
  if((VerboseMask & VERBOSE_CONFIG) != 0){
    fprintf(stdout, "Total simulation time = %lf ms\n", Simulation->TotalTimeSim);
    fprintf(stdout, "Time step             = %e ms\n", Simulation->TimeStep);
    fprintf(stdout, "Nb threads            = %d\n", Simulation->NbThreads);
    fprintf(stdout, "Nb models             = %d\n", Simulation->NumberNeuralModels);
    for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
      fprintf(stdout, " -> Model %2d : Nb neurons = %d\n", ModelNum, Simulation->NeuralModels[ModelNum].NumberOfNeurons);
    }
    for(int SrcModel=0; SrcModel<Simulation->NumberNeuralModels; SrcModel++){
      for(int DstModel=0; DstModel<Simulation->NumberNeuralModels; DstModel++){
        fprintf(stdout, "Synaps Model %2d %2d : Nb inputs = %d\n", SrcModel, DstModel, Simulation->SynapseModels[SrcModel][DstModel].NumberOfConnections);
      }
    }
    fprintf(stdout, "Init OK: %d threads.\n", Simulation->NbThreads);
    fprintf(stderr, "Init OK: %d threads.\n", Simulation->NbThreads);
  }
  
  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_Init, Simulation->Time_Init);
}
