/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Main inlcude file of the SiReNe Software.
 *
 * Data structures and main functions to simulate biological neuronal networks
 * with adjustable properties (normal state, Parkinson's disease,...).
 ******************************************************************************/

#ifndef SIRENE_H
#define SIRENE_H

/*******************************************************************************
 * Standard libraries
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/time.h>
#include <vector>

/*******************************************************************************
 * Internal libraries
 ******************************************************************************/
#include "GenLib.h"

/*******************************************************************************
 * Macros definitions for compiling without OpenMP
 ******************************************************************************/
#ifndef _OPENMP
#define omp_get_thread_num()  0 ///< Alias to OpenMP function returning the thread number
#define omp_get_num_threads() 1 ///< Alias to OpenMP function returning the number of threads
inline double omp_get_wtime() {struct timeval tv; gettimeofday(&tv, NULL); return(tv.tv_sec + 1e-6*tv.tv_usec);} ///< Replacement of OpenMP timer function when OpenMP is not used
#else
#include <omp.h>
#endif

/*******************************************************************************
 * Flags and constants
 ******************************************************************************/

// Flags for the different pieces of information to display
#define QUIET             0       ///< Flag for minimal display in initialization
#define VERBOSE_CONFIG    1       ///< Flag for configuration display in initialization
#define VERBOSE_DETAILS   2       ///< Flag for detailed network display in initialization
#define VERBOSE_TIME      4       ///< Flag for execution time display after simulation

// Flags for the different variables to save in file
#define NO_RECORD              0  ///< 0000000B flag value for no saving during the simulation
#define SAVE_POTENTIALS        1  ///< 0000001B flag value for saving the potentials during the simulation
#define SAVE_VARIABLES         2  ///< 0000010B flag value for saving the variables during the simulation
#define SAVE_INPUT_CURRENTS    4  ///< 0000100B flag value for saving the input currents during the simulation
#define SAVE_SYNAPTIC_CURRENTS 8  ///< 0001000B flag value for saving the synaptic currents during the simulation
#define SAVE_LFP               16 ///< 0010000B flag value for saving the LFP during the simulation
#define SAVE_SPIKES            32 ///< 0100000B flag value for saving the spike times and locations during the simulation
#define SAVE_NBSPIKES          64 ///< 1000000B flag value for saving the number of spikes during the simulation

// Constants for memory and threads management
#define BUFFER_SIZE       1000000 ///< Spikes buffer size
#define NBMEMSTEP         2       ///< Number of stored time steps
#define MAXTHREADS        128     ///< Max number of threads

/*******************************************************************************
 * Definitions of enumerations for the different types of functions
 * used in the models
 ******************************************************************************/

/// Synaptic current types: exponential model wo/w noise (may be extended with other types of synaptic currents)
typedef enum {EXP_CURRENT=0, EXP_CURRENT_WITH_NOISE} SynapticCurrentType;
/// Spike interpolation type: NO_INTERP=Spike at time step, LINEAR_INTERP=Lines, BEZIER_INTERP=Quadratic Bézier curve
typedef enum {NO_INTERP=0, LINEAR_INTERP, BEZIER_INTERP} InterpolatorType;
/// Declaration of the simulation data structure
typedef struct Simul Simul;

/*!*****************************************************************************
 * @brief Description of a spike event
 ******************************************************************************/
typedef struct
{
  double  Time;         ///< Spike time
  long    NeuronNumber; ///< Index of the spiking neuron
} SpikeEvent;

/*!*****************************************************************************
 * @brief Description of a neuron
 ******************************************************************************/
typedef struct
{
  double     *InputParameters;      ///< Input current parameters 
  double     *InitialValues;        ///< Initial values of the neural model (for example: V, m, n, h)
  double      (*InputType)(Simul *Simulation, int ModelNum, int NeuronNum, double t);
                                    ///< Input current function
  int         PotentialIndex;       ///< Potential index for the Variables array stored in the Simul structure
  int         NumberOfSpikes;       ///< Number of spikes for each neuron
  int         NumberOfParameters;   ///< Number of parameters
  uint        GlobalNumNeuron;      ///< Global Id of the neuron
  PRNGState   RandomState;          ///< State of the pseudo-random number generator associated to the neuron
  double     *Parameters;           ///< Parameters array
  double      InputCurrent;         ///< Input Current
  double    **SynCurrentFactor;     ///< Synaptic current factor after spike detection
  double      SpikeTreshold;        ///< Potential threshold used for spike detection
  double      SynapticCurrent;      ///< Synaptic current
  double      LastSpikeTime;        ///< Last spike time of the neuron
  double      dVMidStep[NBMEMSTEP]; ///< Potential derivative at mid time-step
  double      VMidStep[NBMEMSTEP];  ///< Potential at mid time-step
} Neuron;

/*!*****************************************************************************
 * @brief Description of a neural model containing neurons of the same type
 ******************************************************************************/
typedef struct
{
  Neuron *Neurons;            ///< List of neurons in the model
  int     NumberOfNeurons;    ///< Number of neurons
  int     NumberOfParameters; ///< Number of parameters
  int     NumberOfVariables;  ///< Number of variables
  int     FirstNeuronNumber;  ///< First neuron number in global list
  double  SpikeTreshold;      ///< Spike detection by a potential threshold
  double *Parameters;         ///< Parameters array

  void  (*ODEModel)(Simul *Simulation, int ModelNum, int NeuronNum, double t, double* Variables, double* Derivatives);
                              ///< ODE of the neural model
} NeuralModel;

/*!*****************************************************************************
 * @brief Description of a synaptic model
 ******************************************************************************/
typedef struct
{
  SynapticCurrentType  CurrentType;         ///< Synaptic current type
  int                  NumberOfParameters;  ///< Number of parameters (-1 when there is no connection between the models)
  int                  NumberOfVariables;   ///< Number of variables 
  int                  NumberOfConnections; ///< Number of connections between the two models
  double               Conductance;         ///< Synaptic conductance
  double              *Parameters;          ///< Parameters array 
} SynapseModel;

/*!*****************************************************************************
 * @brief Description of an entire simulation
 ******************************************************************************/
struct Simul
{
  // Verbose
  ushort VerboseMask;                    ///< Mask flag indicating the information to display

  // Models
  NeuralModel   *NeuralModels;           ///< List of neural models in the simulation
  SynapseModel **SynapseModels;          ///< 2D (Src -> Dst) array of synaptic models between the neural models
  int            NumberNeuralModels;     ///< Number of neural models
  int            NumberSynapseModels;    ///< Number of synapse models
  int            TotalNumberOfNeurons;   ///< Total number of neurons
  int            NumberOfVariables;      ///< Number of variables
  double         sLFP;                   ///< Sum for each neuron of the LFP

  // Computation types
  InterpolatorType InterpolationType;    ///< Interpolation type

  // Simulated time management
  double  Start;                         ///< Simulation start
  double  TimeStep;                      ///< Time step
  double  TotalTimeSim;                  ///< Simulation execution time 
  double  t[NBMEMSTEP];                  ///< Simulated Time
  double *Variables[NBMEMSTEP];          ///< Variables array (in 2D) containing NBMEMSTEP consecutive time-steps
  double *DVar[NBMEMSTEP];               ///< Variables derivatives array (in 2D) containing NBMEMSTEP consecutive time-steps
  int     CrtNum;                        ///< Index of current time
  int     PrvNum;                        ///< Index of previous time

  // Filenames and data saving management
  char    VariablesFileName[256];        ///< Variables file name
  char    PotentialsFileName[256];       ///< Potentials file name
  char    InputCurrentsFileName[256];    ///< Input Current file name
  char    SynapticCurrentsFileName[256]; ///< Synaptic currents file name
  char    SpikesFileName[256];           ///< Spike times file name
  char    NbSpikesFileName[256];         ///< Number of spikes for each neuron file name
  char    LFPFileName[256];              ///< LFP file name
  ushort  SaveMask;                      ///< Mask flag indicating the variables to save
  int     ResultsInterval;               ///< Number of iterations between results display and/or saving
  
  // Generic functions
  void (*RandGenerConnect)(Simul *Simulation, int ModelNum, int NeuronNum, double SpikeTime);
                                         ///< Pseudo-random connectivity generator function
  void (*NumericalMethod)(Simul *Simulation, 
                          void (*Update)(Simul *Simulation, double tPrev, double t, double *Var, double *Der),
                          double *TmpVar, double *TmpDer);
                                         ///< Numerical method function

  // Spikes management
  FILE       *SpikesFile;                ///< Spikes file
  SpikeEvent *BufferSpikes;              ///< Spikes buffer
  int         NbSpInBuf;                 ///< Nb of spikes stored in the buffer
  
  // Multi-threading management
  int     NbThreads;                     ///< Number of threads

  // Timers
  double  ChronoStart;                        ///< Start timer
  double  Chrono_UpdateModels[MAXTHREADS];    ///< UpdateModels timer
  double  Chrono_UpdateNeuron[MAXTHREADS];    ///< UpdateNeuron timer
  double  Chrono_NumericalMethod[MAXTHREADS]; ///< NumericalMethod timer
  double  Chrono_SpikeProcessing[MAXTHREADS]; ///< SpikeProcessing timer
  double  Chrono_Iter[MAXTHREADS];            ///< Iter timer
  double  Chrono_Advance[MAXTHREADS];         ///< Advance timer
  double  Chrono_Init[MAXTHREADS];            ///< Init timer
  double  Chrono_Detection[MAXTHREADS];       ///< Detection timer
  double  Chrono_RandGenConn[MAXTHREADS];     ///< RandomConnections timer

  // Execution times
  double  Time_UpdateModels[MAXTHREADS];      ///< Elapsed time in UpdateModels
  double  Time_UpdateNeuron[MAXTHREADS];      ///< Elapsed time in UpdateNeuron
  double  Time_NumericalMethod[MAXTHREADS];   ///< Elapsed time in NumericalMethod
  double  Time_SpikeProcessing[MAXTHREADS];   ///< Elapsed time in SpikeProcessing
  double  Time_Iter[MAXTHREADS];              ///< Elapsed time in Iter
  double  Time_Advance[MAXTHREADS];           ///< Elapsed time in Advance
  double  Time_Init[MAXTHREADS];              ///< Elapsed time in Init
  double  Time_Detection[MAXTHREADS];         ///< Elapsed time in Detection
  double  Time_RandGenConn[MAXTHREADS];       ///< Elapsed time in RandomConnections
};

/*******************************************************************************
 * SiReNe functions
 ******************************************************************************/

/*!*****************************************************************************
 * @brief Computation of one time-step from (t) to (t + dt)
 *
 * Network state is computed at (t + dt) using a numerical method (Euler or Runge-Kutta 2).\n
 * Detection of spikes using the SpikeDetection() function.
 *
 * @param Simulation Structure where all simulation information is stored
 * @param TmpVar     Variables array for intermediate time-steps computed in the numerical method
 * @param TmpDer     Derivatives array for intermediate time-steps computed in the numerical method
 ******************************************************************************/
extern void Iter(Simul *Simulation, double *TmpVar, double *TmpDer);

/*!*****************************************************************************
 * @brief Simulation process over time-steps
 *
 * Iterative process over the time-steps:
 *   - Calls to the Iter() function to solve the ODE system of the network at each time-step
 * 
 * Possibility to save data like VARIABLES, POTENTIALS, INPUT_CURRENTS, SYNAPTIC_CURRENTS, LFP, SPIKES and NBSPIKES.\n
 * All files are saved in binary mode.
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
extern void Advance(Simul *Simulation);

/*!*****************************************************************************
 * @brief Initialization of the simulation structure
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
extern void InitSimu(Simul *Simulation);

/*!*****************************************************************************
 * @brief Preparation of simulation data according to the neural network description
 *
 * Initialization of:
 *   - Simulation structure:
 *     - SaveMask for data saving during the simulation
 *     - Time index PrvNum (previous time-step)
 *     - Crt (current time-step)
 *     - Time t0 = simulation starting point 
 *     - Total number of neurons
 *     - Number of variables 
 *     - Allocation of the variables array
 *     - Allocation of the derivatives of the variables array
 *     - Sum of the LFP
 *   - Neuron structure:
 *     - Input/Synaptic current
 *     - Global number neuron
 *     - Number of parameters
 *     - Parameters
 *     - Synaptic current factor
 *     - Spike threshold
 *     - Potential index 
 *     - Mid-time step potential 
 *     - Mid-time step derivative of the potential
 *     - Number of spikes
 *     - Last spike time
 *   - Message:
 *     - Display of initialization information
 *       - Initialization display is activated via the VerboseMask parameter
 *   - Chrono:
 *     - Time initalization for each function
 *
 * @param Simulation      Structure where all simulation information is stored
 * @param VerboseMask     Mask of information to display
 * @param SaveMask        Mask of variables to save to file
 * @param ResultsInterval Number of iterations between results display / save
 ******************************************************************************/
extern void PrepareSimu(Simul *Simulation, ushort VerboseMask, ushort SaveMask, int ResultsInterval);

/*******************************************************************************
 * Macros definitions for the functions pre and post-processing to measure
 * the execution times
 ******************************************************************************/
#ifdef TIME_CALLS
#define INPUT_CALL(chrono) { chrono[omp_get_thread_num()] = omp_get_wtime(); }                                     ///< Macro to be placed at the beginning of each function requiring execution time profiling
#define RETURN_CALL(chrono,cumul) { int num = omp_get_thread_num(); cumul[num] += omp_get_wtime() - chrono[num]; } ///< Macro to be placed at the end of each function requiring execution time profiling
#else
#define INPUT_CALL(chrono)        ///< Macro to be placed at the beginning of each function requiring execution time profiling (empty macro when time profiling is not activated)
#define RETURN_CALL(chrono,cumul) ///< Macro to be placed at the end of each function requiring execution time profiling (empty macro when time profiling is not activated)
#endif

#endif
