/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Spike processing functions
 ******************************************************************************/

#include "Spikes.h"

/*!*****************************************************************************
 * @brief Spike processing of neuron NeuronNum in model ModelNum at current time
 *
 * The following operations are performed to propagate a spike to post-synaptic neurons:
 *   - Pseudo-random connectivity generation
 *   - Update of synaptic currents of post-synaptic neurons
 *   - Spike time and neuron number are stored in memory
 *   - Spike time and neuron number are saved in a file (if activated in the settings)
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param SpikeTime  Time of the last spike of neuron NeuronNum in model ModelNum
 ******************************************************************************/
void SpikeProcessing(Simul *Simulation, int ModelNum, int NeuronNum, double SpikeTime)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_SpikeProcessing);
  
  //
  // Display of spike time and neuron number if corresponding compilation setting is set to yes: one concurrent thread
  //
  #ifdef DISPLAYSPIKE
  fprintf(stdout, "%.20f \t %5u\n", SpikeTime, Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron);
  fflush(stdout);
  #endif

  //
  // Spikes are saved in a binary file: one thread at a time
  //
  if((Simulation->SaveMask & SAVE_SPIKES) != 0) {
                         //
    #pragma omp critical // Mutual exclusion (one thread at a time) to ensure coherency
                         //
    {
      /*! @warning
       *  For very large networks, the number of spikes may be very large, generating huge spikes file\n
       *  => An additional test should be added around the spike saving, in order to store and save only a subset of neurons
       */

      // Storage of current spike in the buffer
      Simulation->BufferSpikes[Simulation->NbSpInBuf].NeuronNumber = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].GlobalNumNeuron;
      Simulation->BufferSpikes[Simulation->NbSpInBuf].Time         = SpikeTime;
      Simulation->NbSpInBuf++;
      
      // Saving to file when the buffer is full
      if(Simulation->NbSpInBuf == BUFFER_SIZE){
        fwrite(Simulation->BufferSpikes, sizeof(SpikeEvent), BUFFER_SIZE, Simulation->SpikesFile);
        fflush(Simulation->SpikesFile);
        Simulation->NbSpInBuf = 0;
      }
    }
  }
    
  // Update of the number of spikes for neuron NeuronNum in model ModelNum: one concurrent thread
  Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].NumberOfSpikes++;
  
  //
  // Random generation of post-synaptic connection and updating of synaptic current of post-synaptic neurons of neuron NeuronNum in model ModelNum: one concurrent thread
  //
  (*Simulation->RandGenerConnect)(Simulation, ModelNum, NeuronNum, SpikeTime);

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_SpikeProcessing, Simulation->Time_SpikeProcessing);
}

/*!*****************************************************************************
 * @brief Spikes detection in current time step 
 *
 * For each spike, the spike time is determined according to the specified interpolation method:
 *   - Threshold:
 *     - Spike time set at the time step
 *   - Linear:
 *     - Intersection of two lines determined by the derivatives of the potential at times (t) and (t+dt)
 *   - Quadratic Bezier curve:
 *     - Extremities of the Bézier curve are determinated by the potential at times (t) and (t+dt)
 *     - Tangents control point of the Bézier curve is deduced from the intersection of the two lines
 *     - The two lines are determined by the derivatives of the potential at times (t) and (t+dt)
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
void SpikeDetection(Simul *Simulation) 
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_Detection);

  //
  // Spike detection for each model of neurons: all threads
  //
  for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){

    //
    // Spike detection for each neuron in current model: parallel threads
    //
    #pragma omp for schedule(dynamic, 10) /// @remark The size of parallel iterations blocks of the neurons loop should be tuned according to the machine
    for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons ; NeuronNum++){

      int v;                                      // Potential Index of the current neuron in the global variables array
      double PreviousPotential, CurrentPotential; // Current and previous potentials of current neuron
      double SpikeTime =  -1.00000e+308;          // Spike time

      v                 =   Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].PotentialIndex;
      PreviousPotential =   Simulation->Variables[Simulation->PrvNum][v];
      CurrentPotential  =   Simulation->Variables[Simulation->CrtNum][v];

      //
      // Spike detection using the threshold and the potential's derivatives at (t) and (t+dt)
      //
      if(((PreviousPotential > Simulation->NeuralModels[ModelNum].SpikeTreshold) || (CurrentPotential > Simulation->NeuralModels[ModelNum].SpikeTreshold))
         && (Simulation->DVar[Simulation->PrvNum][v] > 0.0) 
         && (Simulation->DVar[Simulation->CrtNum][v] < 0.0)){
  
        // Interpolation method
        switch(Simulation->InterpolationType){

        case NO_INTERP: // No interpolation: spike aligned to the time step
          SpikeTime = Simulation->t[Simulation->CrtNum];
          break;

        case LINEAR_INTERP: // Linear interpolation: intersection of the two tangents at the time steps that surround the spike
          {
            // Line D_(t) passing throught V(t) with slope dV(t)/dt: y = ax + b <=> V(t) = dV(t)/dt * t + b0
            // Intersection point b0 between the line and the y-axis
            double b0 = PreviousPotential - Simulation->DVar[Simulation->PrvNum][v] * Simulation->t[Simulation->PrvNum];

            // Line D_(t+dt) passing throught V(t+dt) with slope dV(t+dt)/dt: y = ax + b <=> V(t+dt) = dV(t+dt)/dt * t + b1
            // Intersection point b1 between the line and the y-axis
            double b1 = CurrentPotential - Simulation->DVar[Simulation->CrtNum][v] * Simulation->t[Simulation->CrtNum];

            // Time at mid-step: t+dt/2
            double tMidStep = Simulation->t[Simulation->PrvNum] + 0.5 * Simulation->TimeStep;
   
            // Spike time computation
            SpikeTime = (b1- b0) / ( Simulation->DVar[Simulation->PrvNum][v] - Simulation->DVar[Simulation->CrtNum][v]);

            // Spike time set to (t+0.5dt) when the interpolation results in a smaller potential
            if(   (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].dVMidStep[Simulation->PrvNum] >= 0 && SpikeTime < tMidStep)
                  || (Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].dVMidStep[Simulation->PrvNum] <= 0 && SpikeTime > tMidStep)){
              SpikeTime = Simulation->t[Simulation->PrvNum] + 0.5 * Simulation->TimeStep;
            }
          }
          break;

        case BEZIER_INTERP: // Second order Bezier curve interpolation
          {
            // Line D_(t) passing throught V(t) with slope dV(t)/dt: y = ax + b <=> V(t) = dV(t)/dt * t + b0
            // Intersection point b0 between the line and the y-axis
            double b0 = PreviousPotential - Simulation->DVar[Simulation->PrvNum][v] * Simulation->t[Simulation->PrvNum];

            // Line D_(t+dt) passing throught V(t+dt) with slope dV(t+dt)/dt: y = ax + b <=> V(t+dt) = dV(t+dt)/dt * t + b1
            // Intersection point b1 between the line and the y-axis
            double b1 = CurrentPotential - Simulation->DVar[Simulation->CrtNum][v] * Simulation->t[Simulation->CrtNum];

            // Second control point P2(V2,t2) of the Bézier curve deduced from the intersection of the two lines D_(t) and D_(t+dt) 
            double DV0mDV1 = Simulation->DVar[Simulation->PrvNum][v] - Simulation->DVar[Simulation->CrtNum][v];
            double t2 = (b1 - b0) / DV0mDV1;
            double V2 = (b1 * Simulation->DVar[Simulation->PrvNum][v] - b0 * Simulation->DVar[Simulation->CrtNum][v]) / DV0mDV1;

            // Construction of the three control points (P1, P2, P3) of the Bézier curve and the spike point SP
            Point2D P1, P2, P3, SP; 
            P1.x = Simulation->t[Simulation->PrvNum];
            P2.x = t2;
            P3.x = Simulation->t[Simulation->CrtNum];

            P1.y = PreviousPotential;
            P2.y = V2;
            P3.y = CurrentPotential;

            // Computation of the maximum point of the Bezier curve along y axis
            SP = BezierMax('y', &P1, &P2, &P3);

            SpikeTime = SP.x;
          }
          break;

        } // End switch

          //
          // Spike processing: one concurrent thread
          //
        SpikeProcessing(Simulation, ModelNum, NeuronNum, SpikeTime);
  
        // Storage of the last spike time
        Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].LastSpikeTime = SpikeTime;
  
      } // End if spike detection
    } // End for neurons
  } // End for models

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_Detection, Simulation->Time_Detection);
}
