/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Spike processing functions
 ******************************************************************************/

#ifndef SPIKES_H
#define SPIKES_H

#include "Sirene.h"
#include "BezierInterpolation.h"

/*!*****************************************************************************
 * @brief Spike processing of neuron NeuronNum in model ModelNum at current time
 *
 * The following operations are performed to propagate a spike to post-synaptic neurons:
 *   - Pseudo-random connectivity generation
 *   - Update of synaptic currents of post-synaptic neurons
 *   - Spike time and neuron number are stored in memory
 *   - Spike time and neuron number are saved in a file (if activated in the settings)
 *
 * @param Simulation Structure where all simulation information is stored
 * @param ModelNum   Neural model number
 * @param NeuronNum  Neuron number
 * @param SpikeTime  Time of the last spike of neuron NeuronNum in model ModelNum
 ******************************************************************************/
extern void SpikeProcessing(Simul *Simulation, int ModelNum, int NeuronNum, double SpikeTime);

/*!*****************************************************************************
 * @brief Spikes detection in current time step 
 *
 * For each spike, the spike time is determined according to the specified interpolation method:
 *   - Threshold:
 *     - Spike time set at the time step
 *   - Linear:
 *     - Intersection of two lines determined by the derivatives of the potential at times (t) and (t+dt)
 *   - Quadratic Bezier curve:
 *     - Extremities of the Bézier curve are determinated by the potential at times (t) and (t+dt)
 *     - Tangents control point of the Bézier curve is deduced from the intersection of the two lines
 *     - The two lines are determined by the derivatives of the potential at times (t) and (t+dt)
 *
 * @param Simulation Structure where all simulation information is stored
 ******************************************************************************/
extern void SpikeDetection(Simul *Simulation);

#endif
