/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Synaptic current model.
 *
 * Synapse model:
 *   - Single exponential conductance:
 *     - "Spiking Neuron Models", Wulfram Gerstner and Werner M. Kistler,\n
 *       Cambridge University Press, 2002, pp 62-63 (2.4.2)
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "Sirene.h"
#include "SynapticCurrent.h"

/*!*****************************************************************************
 * @brief Synaptic current with exponential conductance at current time (t) from previous time (tPrev)
 *
 * @param Simulation        Structure where all simulation information is stored
 * @param ModelNum          Neural model number
 * @param NeuronNum         Neuron number
 * @param PreSynModel       Pre-synaptic neural model number
 * @param tPrev             Previous time from which the synaptic current is computed
 * @param t                 Current time at which the synaptic current is computed
 * @param CurrPostPotential Current post-potential
 *
 * @result Synaptic current
 ******************************************************************************/
double SynapticCurrent(Simul *Simulation, int ModelNum, int NeuronNum, int PreSynModel, double tPrev, double t, double CurrPostPotential)
{
  double Current = 0.0;
  
  if(Simulation->SynapseModels[PreSynModel][ModelNum].NumberOfParameters >= 0){                 // Test if the synapses from PreSynModel to ModelNum exist
    double SynConductance = Simulation->SynapseModels[PreSynModel][ModelNum].Conductance;       // Conductance between the two neural models
    double DecayTime = Simulation->SynapseModels[PreSynModel][ModelNum].Parameters[0];          // Time constant of decay
    double PotentialInversion = Simulation->SynapseModels[PreSynModel][ModelNum].Parameters[1]; // Reversal potential

    // Selection of the synaptic current model
    switch(Simulation->SynapseModels[PreSynModel][ModelNum].CurrentType){
    case EXP_CURRENT: // Exponential synaptic current model
      {
        double sumSECF = 0.0;                                                    // Sum of synaptic exponential current factors over the threads
        double Fact = exp(-(t - tPrev) / DecayTime);                             // Second part of the exponential decay factor (between the beginning of current time step and the current time)

        for(int numT=0; numT<Simulation->NbThreads; ++numT){                     // Aggregation of all partial current factors distributed over the threads
          Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[PreSynModel][numT] *= Fact;    // Update of each partial factor 
          sumSECF += Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[PreSynModel][numT]; // Contribution of each thread to the synaptic current factor
        }

        Current = (CurrPostPotential - PotentialInversion)                       // Final input current
                  * sumSECF * SynConductance;
      }
      break;
    case EXP_CURRENT_WITH_NOISE: // Exponential synaptic current model with noise
      {
        double Amp = 0.71;                                                       // Noise amplitude 
        double sumSECF = 0.0;                                                    // Sum of synaptic exponential current factors over the threads
        double Fact = exp(-(t - tPrev) / DecayTime);                             // Second part of the exponential decay factor (between the beginning of current time step and the current time)

        for(int numT=0; numT<Simulation->NbThreads; ++numT){                     // Aggregation of all partial current factors distributed over the threads
          Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[PreSynModel][numT] *= Fact;    // Update of each partial factor 
          sumSECF += Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynCurrentFactor[PreSynModel][numT]; // Contribution of each thread to the synaptic current factor
        }

        Current = (CurrPostPotential - PotentialInversion)                       // Final input current
                  * sumSECF * SynConductance
                  * SpecificNoise(Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].RandomState, Amp);
      }
      break;
    }
  }

  return Current;
}

/*!*****************************************************************************
 * @brief Update of synaptic current of a post-synatpic neuron after spike detection
 *
 * @param Simulation    Structure where all simulation information is stored
 * @param ModelNum      Neural model number
 * @param NeuronNum     Neuron number
 * @param PostSynModel  Post-synaptic neural model number
 * @param PostSynNeuron Post-synaptic neuron to update
 * @param SpikeTime     Spike time
 ******************************************************************************/
void SynapticSpike(Simul *Simulation, int ModelNum, int NeuronNum, int PostSynModel, int PostSynNeuron, double SpikeTime)
{
  double DecayTime = Simulation->SynapseModels[ModelNum][PostSynModel].Parameters[0]; // Time constant of decay

  // Selection of the synaptic current model
  switch(Simulation->SynapseModels[ModelNum][PostSynModel].CurrentType){
  case EXP_CURRENT:            // Exponentiel synaptic current model
  case EXP_CURRENT_WITH_NOISE: // Exponentiel synaptic current model with noise
                               // Both have the same computation at this point but differ in SynapticCurrent (applied at next time step)
    {
      // Application of the first part of the exponential decay factor (between the spike time and the end of the current time step)
      Simulation->NeuralModels[PostSynModel].Neurons[PostSynNeuron].SynCurrentFactor[ModelNum][omp_get_thread_num()] += exp(-(Simulation->t[Simulation->CrtNum] - SpikeTime) / DecayTime);
    }
    break;
  }
}
