/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief MSN network example with Parkinson's disease and normal conditions.
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "SireneComponents.h"

/*!*****************************************************************************
 * @brief Command line help function
 ******************************************************************************/
void help(char *prog)
{
  printf("%s -c <int> -d <float> -h -i <char> -n <int> -r -s <float> -s <int> -t <int>\n", prog);
  printf("   -c <int>   : number of connections per neuron\n");
  printf("   -d <float> : simulation run time in ms\n");
  printf("   -h         : this help\n");
  printf("   -i <char>  : type of input current ('c': constant 'n': noisy \n");
  printf("   -n <int>   : number of neurons\n");
  printf("   -s <float> : time step in ms\n");
  printf("   -t <int>   : number of threads\n");
  exit(EXIT_SUCCESS);
}

/*!*****************************************************************************
 * @brief Main function of the simulation program
 ******************************************************************************/
int main(int argc, char *argv[])
{
  Simul Simulation;                          // Simulation structure
  InterpolatorType Interpol = BEZIER_INTERP; // Interpolation type
  int NbNeurons             = 2000;          // Number of neurons in the model
  double ConnectionRate     = 1;             // Rate (%) of connections per neuron
  int NbThreads             = 1;             // Number of threads
  int ResultsPeriod         = 1e8;           // Period of results display / saving
  double ActPotThreshold    = 0.0;           // Action potential threshold in mV
  double TimeStep           = 5e-3;          // Time step in ms
  double RunTime            = 4000.0;        // Simulation run time in ms
  double AmpliInitVal       = 0.1;           // Noise amplitude for the initial values
  char InputCurrentType     = 'c';           // Type of neurons input current ('c': constant, 'n': noisy current)
  int opt;                                   // Command line options 
 
  /*****************************************************************************
   * Parameters
   ****************************************************************************/
  while ((opt = getopt(argc, argv, "c:d:hi:n:rs:t:")) != -1) {
    switch (opt) {
    case 'c':                   // Connection rate of the neurones
      ConnectionRate = atof(optarg);
      break;
    case 'd':                   // Simulation run time
      RunTime = atof(optarg);
      break;
    case 'h':                   // Command line help
      help(argv[0]);
      break;
    case 'i':                   // Input current type
      InputCurrentType = optarg[0];
      break;
    case 'n':                   // Number of neurons
      NbNeurons = atoi(optarg);
      break;
    case 's':                   // Time step
      TimeStep = atof(optarg);  /// @warning The time step must not be greater than 0.03
      break;
    case 't':                   // Number of threads
      NbThreads = atoi(optarg);
      break;
    }
  }

  /*****************************************************************************
   * Simulation Init
   ****************************************************************************/
  InitSimu(&Simulation);                            // Initialization of the simulation structure (all values set to 0)
  Simulation.Start             = 0.0;               // Simulation start in ms
  Simulation.TotalTimeSim      = RunTime;           // Simulation run time in ms
  Simulation.TimeStep          = TimeStep;          // Time step in ms
  Simulation.InterpolationType = Interpol;          // Interpolation type
  Simulation.NumericalMethod   = RungeKutta2;       // Numerical method: Euler/RK2
  Simulation.RandGenerConnect  = RandomConnections; // Pseudo-random connectivity generator
  Simulation.NbThreads         = NbThreads;         // Number of threads
  
  /*****************************************************************************
   * Backup filenames
   ****************************************************************************/
  strcpy(Simulation.VariablesFileName,        "Variables.bin");
  strcpy(Simulation.PotentialsFileName,       "Potentials.bin");
  strcpy(Simulation.InputCurrentsFileName,    "InputCurrents.bin");
  strcpy(Simulation.SynapticCurrentsFileName, "SynapticCurrents.bin");
  strcpy(Simulation.SpikesFileName,           "Spikes.bin");
  strcpy(Simulation.LFPFileName,              "LFP.bin");
  strcpy(Simulation.NbSpikesFileName,         "NbSpikes.bin");

  /*****************************************************************************
   * Neural Models allocation
   ****************************************************************************/
  Simulation.NumberNeuralModels = 1;
  Simulation.NeuralModels = (NeuralModel*)MYMALLOC(Simulation.NumberNeuralModels*sizeof(NeuralModel));

  /*****************************************************************************
   * MSN Model
   ****************************************************************************/
  Simulation.NeuralModels[0].ODEModel           = MSN;
  Simulation.NeuralModels[0].NumberOfNeurons    = NbNeurons;
  Simulation.NeuralModels[0].FirstNeuronNumber  = 0;
  Simulation.NeuralModels[0].SpikeTreshold      = ActPotThreshold;
  Simulation.NeuralModels[0].NumberOfVariables  = 5;

  /*****************************************************************************
   * Maximal conductances
   ****************************************************************************/
  Simulation.NeuralModels[0].NumberOfParameters = 4;
  Simulation.NeuralModels[0].Parameters         = (double *) MYMALLOC(Simulation.NeuralModels[0].NumberOfParameters*sizeof(double));
  Simulation.NeuralModels[0].Parameters[0]      = 0.1;   // Leak current conductance (in mS)
  Simulation.NeuralModels[0].Parameters[1]      = 100.0; // Na current conductance (in mS)
  Simulation.NeuralModels[0].Parameters[2]      = 80.0;  // K current conductance (in mS)
  Simulation.NeuralModels[0].Parameters[3]      = 0.25;  // M current conductance (in mS) (normal conditions = 1.34, Parkinson conditions = 0.25)

  /*****************************************************************************
   * Neurons allocation
   ****************************************************************************/
  for(int ModelNum = 0; ModelNum < Simulation.NumberNeuralModels; ModelNum++){
    Simulation.NeuralModels[ModelNum].Neurons = (Neuron *)MYMALLOC(Simulation.NeuralModels[ModelNum].NumberOfNeurons*sizeof(Neuron));
  }

  /*****************************************************************************
   * Neurons settings
   ****************************************************************************/
  for(int NeuronNum = 0; NeuronNum < Simulation.NeuralModels[0].NumberOfNeurons; NeuronNum++){
    // PRNG state initialization
    InitPRNG(NeuronNum, &Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState);

    // Variables initialization:
    // -> Each neuron parameter is initialized with a mean value and a noise
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues    = (double *)MYMALLOC(Simulation.NeuralModels[0].NumberOfVariables*sizeof(double));
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues[0] = -63.826  * SpecificNoise(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState, AmpliInitVal); // Initial membrane potential of the neuron (in mV)
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues[1] = 0.027781 * SpecificNoise(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState, AmpliInitVal); // Activation gating variable of Na current (in 1)
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues[2] = 0.990915 * SpecificNoise(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState, AmpliInitVal); // Inactivation gating variable of Na current (in 1)
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues[3] = 0.062045 * SpecificNoise(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState, AmpliInitVal); // Activation gating variable of K current (in 1)
    Simulation.NeuralModels[0].Neurons[NeuronNum].InitialValues[4] = 0.022788 * SpecificNoise(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState, AmpliInitVal); // Activation gating variable of M current (in 1)

    // Input Current
    switch(InputCurrentType){
    case 'c': // Constant current
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputType          = ConstantInputCurrent;
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters    = (double *)MYMALLOC(sizeof(double));   // Parameters array allocation
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters[0] = 0.57;                                 // Input current
      break;
    case 'n': // Noisy step current
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputType          = StepwiseInputCurrentWithNoise;
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters    = (double *)MYMALLOC(4*sizeof(double)); // Parameters array allocation
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters[0] = -10.0;                                // Input current before the step time
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters[1] =   0.5;                                // Input current after the step time
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters[2] = 100.0 * RandomReal(Simulation.NeuralModels[0].Neurons[NeuronNum].RandomState); // Step time
      Simulation.NeuralModels[0].Neurons[NeuronNum].InputParameters[3] =   3.0;                                // Amplitude of the noise
      break;
    }
  }

  /*****************************************************************************
   * Synapse Models allocation
   * -> The maximal number of synapse models is (number of neural models)^2
   ****************************************************************************/
  Simulation.NumberSynapseModels = Simulation.NumberNeuralModels * Simulation.NumberNeuralModels;
  Simulation.SynapseModels       = (SynapseModel **) SynapseModelsAlloc(Simulation.NumberSynapseModels);

  /*****************************************************************************
   * MSN-MSN synapse type
   ****************************************************************************/
  Simulation.SynapseModels[0][0].CurrentType         = EXP_CURRENT_WITH_NOISE;  // Synaptic current type
  Simulation.SynapseModels[0][0].NumberOfVariables   = 0;            // Number of variables
  Simulation.SynapseModels[0][0].NumberOfParameters  = 2;            // Number of parameters
  Simulation.SynapseModels[0][0].NumberOfConnections = (int) rint(ConnectionRate * Simulation.NeuralModels[0].NumberOfNeurons); // Number of connections
  Simulation.SynapseModels[0][0].Parameters          = (double *) MYMALLOC(Simulation.SynapseModels[0][0].NumberOfParameters * sizeof(double));
  Simulation.SynapseModels[0][0].Parameters[0]       = 12;           // Synaptic time constant
  Simulation.SynapseModels[0][0].Parameters[1]       = -80.0;        // Reversal potential of GABA synapses
  if(Simulation.SynapseModels[0][0].NumberOfConnections != 0){
    Simulation.SynapseModels[0][0].Conductance = 0.6 / Simulation.SynapseModels[0][0].NumberOfConnections; // synaptic conductance
  } else {
    Simulation.SynapseModels[0][0].Conductance = 0;
  }

  /*****************************************************************************
   * Simulation execution
   ****************************************************************************/
  PrepareSimu(&Simulation, QUIET, NO_RECORD, ResultsPeriod); // Simulation initialization
  Advance(&Simulation);                                      // Simulation start

  /*****************************************************************************
   * Other example:
   *
   * PrepareSimu(&Simulation, (VERBOSE_CONFIG | VERBOSE_TIME), (SAVE_VARIABLES | SAVE_INPUT_CURRENTS | SAVE_SYNAPTIC_CURRENTS | SAVE_POTENTIALS ), ResultsPeriod);
   ****************************************************************************/

  /*****************************************************************************
   * Memory Release
   ****************************************************************************/
  MemoryRelease(&Simulation);

  /*****************************************************************************
   * Memory Statement
   ****************************************************************************/
  DISPLAYMEM();

  /*****************************************************************************
   * Final Display
   ****************************************************************************/
  fprintf(stderr, "End of SiReNe simulation with %d threads.\n", Simulation.NbThreads);

  return 0;
}
