/*******************************************************************************
 * SiReNe software version 1.0
 * Copyright © 2018-2020, INRIA, CNRS, Université de Lorraine. All rights reserved.
 *
 * Authors : N. AZEVEDO CARVALHO, S. CONTASSOT-VIVIER, L. BUHRY and D. MARTINEZ
 * Version : 1.0
 * Years   : 2018-2020
 *
 * Authors : W. RAPHA, E. HUGUES, D. MARTINEZ and Y. BONIFACE
 * Version : 0.1
 * Years   : 2002-2009
 *
 * License : GNU General Public License, version 2
 *
 * This  program is  free  software; you  can  redistribute it  and/or
 * modify it  under the  terms of  the GNU  General Public  License as
 * published by the Free Software  Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 * MERCHANTABILITY or FITNESS  FOR A PARTICULAR PURPOSE.   See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/
/*!
 * @file 
 * @brief Update functions of neurons and models 
 ******************************************************************************/

#include "Updates.h"

/*!*****************************************************************************
 * @brief Update of all models in the network at time (t) from previous time (tPrev)
 *  
 * @param Simulation  Sructure where all simulation information is stored
 * @param tPrev       Previous time from which models are updated
 * @param t           Current time at which models are updated
 * @param Variables   Variables of the neuronal model at time (t)
 * @param Derivatives Derivatives computed at time (t) using the ODE system of the neuronal model
 ******************************************************************************/
void UpdateModels(Simul *Simulation, double tPrev, double t, double *Variables, double *Derivatives)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_UpdateModels);
  
  //
  // State update for each model: all threads
  //
  for(int ModelNum=0; ModelNum<Simulation->NumberNeuralModels; ModelNum++){
    //
    // State update of all neurons in model ModelNum: parallel threads
    //
    #pragma omp for
    for(int NeuronNum=0; NeuronNum<Simulation->NeuralModels[ModelNum].NumberOfNeurons; NeuronNum++){
      int v = Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].PotentialIndex;

      //
      // State update of neuron NeuronNum in model ModelNum: one concurrent thread
      //
      UpdateNeuron(Simulation, ModelNum, NeuronNum, tPrev, t, &Variables[v], &Derivatives[v]);
    }
  }

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_UpdateModels, Simulation->Time_UpdateModels);
}

/*!*****************************************************************************
 * @brief Update of neuron NeuronNum in model ModelNum at time (t) from previous time (tPrev)
 *
 * Update of the following properties of the specified neuron in the specified model:
 *   - Input current
 *   - Synaptic current
 *   - Derivatives
 *  
 * @param Simulation  Sructure where all simulation information is stored
 * @param ModelNum    Neural model number
 * @param NeuronNum   Neuron number
 * @param tPrev       Previous time from which neuron state is updated
 * @param t           Current time at which neuron state is updated
 * @param Variables   Variables of the neuronal model at time (t)
 * @param Derivatives Derivatives computed at time (t) using the ODE system of the neuronal model
 ******************************************************************************/
void UpdateNeuron(Simul *Simulation, int ModelNum, int NeuronNum, double tPrev, double t, double *Variables, double *Derivatives)
{
  // Pre-processing for execution time profiling
  INPUT_CALL(Simulation->Chrono_UpdateNeuron);
  double CurrPostPotential = Variables[0];
  
  // Input current of neuron NeuronNum: one concurrent thread
  Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputCurrent = (*Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].InputType)(Simulation, ModelNum, NeuronNum, t);

  // Synaptic current of neuron NeuronNum: one concurrent thread
  Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent = 0.0;
  for(int i = 0; i < Simulation->NumberNeuralModels; i++){ // Run through incoming synapses model
    Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].SynapticCurrent += SynapticCurrent(Simulation, ModelNum, NeuronNum, i, tPrev, t, CurrPostPotential);
  }

  // Update of neuron derivatives => ODE system of model ModelNum: one concurrent thread
  (*Simulation->NeuralModels[ModelNum].ODEModel)(Simulation, ModelNum, NeuronNum, t, Variables, Derivatives);

  // Detection of current t at the middle of the time step
  if(Simulation->t[Simulation->PrvNum] + 0.25 * Simulation->TimeStep < t && t < Simulation->t[Simulation->CrtNum] - 0.25 * Simulation->TimeStep){
    Simulation->NeuralModels[ModelNum].Neurons[NeuronNum].dVMidStep[Simulation->PrvNum] = Derivatives[0];
  }

  // Post-processing for execution time profiling
  RETURN_CALL(Simulation->Chrono_UpdateNeuron, Simulation->Time_UpdateNeuron);
}
